"""Unit tests for XP24 Action Table Serializer."""

import pytest

from xp.models.actiontable.msactiontable_xp24 import InputAction, Xp24MsActionTable
from xp.models.telegram.input_action_type import InputActionType
from xp.models.telegram.timeparam_type import TimeParam
from xp.services.actiontable.msactiontable_xp24_serializer import (
    Xp24MsActionTableSerializer,
)
from xp.utils.serialization import de_nibbles


class TestXp24MsActionTableSerializer:
    """Test cases for Xp24MsActionTableSerializer."""

    @pytest.fixture
    def sample_action_table(self):
        """Create sample action table for testing."""
        return Xp24MsActionTable(
            input1_action=InputAction(InputActionType.TOGGLE, TimeParam.NONE),
            input2_action=InputAction(InputActionType.ON, TimeParam.T5SEC),
            input3_action=InputAction(InputActionType.LEVELSET, TimeParam.T5SEC),
            input4_action=InputAction(InputActionType.SCENESET, TimeParam.T5SEC),
            mutex12=True,
            mutex34=False,
            mutual_deadtime=Xp24MsActionTable.MS500,
            curtain12=False,
            curtain34=True,
        )

    @pytest.fixture
    def sample_telegrams(self):
        """Create sample telegrams for testing."""
        return [
            "<R0020044989F17DAAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFA>",
            "<R0020044966F17DAAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFB>",
            "<R0020044986F17DAAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFP>",
            "<R0020041824F17DAAAAAAAAAAABACAEAIBACAEAIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFP>",
            "<R0020044964F17DAAAAABAGADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFD>",
        ]

    def test_from_telegrams_invalid_hex_data(self):
        """Test that invalid hex data raises ValueError with non-hexadecimal characters."""
        # This telegram contains non-hex characters that cause from_hex() to fail
        # Based on the debug log: '<R0020044989F17DAAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFA>'
        valid_telegram = (
            "AAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        )

        msactiontable = Xp24MsActionTableSerializer.from_data(valid_telegram)
        assert msactiontable.input1_action.type == InputActionType.TOGGLE
        assert msactiontable.input2_action.type == InputActionType.TOGGLE
        assert msactiontable.input3_action.type == InputActionType.TOGGLE
        assert msactiontable.input4_action.type == InputActionType.TOGGLE

        assert msactiontable.input1_action.param == TimeParam.NONE
        assert msactiontable.input2_action.param == TimeParam.NONE
        assert msactiontable.input3_action.param == TimeParam.NONE
        assert msactiontable.input4_action.param == TimeParam.NONE

        assert not msactiontable.curtain12
        assert not msactiontable.curtain34
        assert not msactiontable.mutex12
        assert not msactiontable.mutex34

    def test_from_telegrams_from_data(self):
        """Test that invalid hex data raises ValueError with non-hexadecimal characters."""
        # This telegram contains non-hex characters that cause from_hex() to fail
        # Based on the debug log: '<R0020044989F17DAAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFA>'
        valid_msactiontable = (
            "AAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        )

        msactiontable = Xp24MsActionTableSerializer.from_data(valid_msactiontable)
        msactiontable_data = Xp24MsActionTableSerializer.to_data(msactiontable)
        assert valid_msactiontable == msactiontable_data

    def test_from_telegrams_invalid_hex_data2(self):
        """Test that invalid hex data raises ValueError with non-hexadecimal characters."""
        # This telegram contains non-hex characters that cause from_hex() to fail
        # Based on the debug log: '<R0020044964F17DAAAAABAGADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAFD>'
        valid_telegram = (
            "AAAAABAGADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        )

        msactiontable = Xp24MsActionTableSerializer.from_data(valid_telegram)
        assert msactiontable.input1_action.type == InputActionType.ON
        assert msactiontable.input2_action.type == InputActionType.TOGGLE
        assert msactiontable.input3_action.type == InputActionType.TOGGLE
        assert msactiontable.input4_action.type == InputActionType.TOGGLE

        assert msactiontable.input1_action.param == TimeParam.T15SEC
        assert msactiontable.input2_action.param == TimeParam.NONE
        assert msactiontable.input3_action.param == TimeParam.NONE
        assert msactiontable.input4_action.param == TimeParam.NONE

        assert not msactiontable.curtain12
        assert not msactiontable.curtain34
        assert not msactiontable.mutex12
        assert not msactiontable.mutex34

    def test_from_telegrams_de_nibble_0(self):
        """Test that invalid hex data raises ValueError with non-hexadecimal characters."""
        nibble = "AA"

        result = de_nibbles(nibble)
        assert bytearray([0]) == result

    def test_from_telegrams_de_nibble_1(self):
        """Test that invalid hex data raises ValueError with non-hexadecimal characters."""
        nibble = "AB"

        result = de_nibbles(nibble)
        assert bytearray([1]) == result

    def test_from_telegrams_de_nibble_01(self):
        """Test that invalid hex data raises ValueError with non-hexadecimal characters."""
        nibble = "AAAB"

        result = de_nibbles(nibble)
        assert bytearray([0, 1]) == result

    def test_from_telegrams_de_nibble_big(self):
        """Test that invalid hex data raises ValueError with non-hexadecimal characters."""
        nibble = "AAAAADAAADAAADAAADAAAAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

        result = de_nibbles(nibble)
        assert (
            bytearray(
                [
                    0,
                    0,
                    3,
                    0,
                    3,
                    0,
                    3,
                    0,
                    3,
                    0,
                    0,
                    0,
                    12,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
            )
            == result
        )



"""Conbus Raw Service for sending raw telegram sequences.

This service handles sending raw telegram strings without prior validation.
"""

import logging
from datetime import datetime
from typing import Callable, Optional

from twisted.internet.posixbase import PosixReactorBase

from xp.models import ConbusClientConfig
from xp.models.conbus.conbus_raw import ConbusRawResponse
from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.services.protocol import ConbusProtocol


class ConbusRawService(ConbusProtocol):
    """
    Service for sending raw telegram sequences to Conbus modules.

    Uses ConbusProtocol to provide raw telegram functionality
    for sending arbitrary telegram strings without validation.
    """

    def __init__(
        self,
        cli_config: ConbusClientConfig,
        reactor: PosixReactorBase,
    ) -> None:
        """Initialize the Conbus raw service.

        Args:
            cli_config: Configuration for Conbus client connection.
            reactor: Twisted reactor for event loop.
        """
        super().__init__(cli_config, reactor)
        self.raw_input: str = ""
        self.progress_callback: Optional[Callable[[str], None]] = None
        self.finish_callback: Optional[Callable[[ConbusRawResponse], None]] = None
        self.service_response: ConbusRawResponse = ConbusRawResponse(
            success=False,
        )
        # Set up logging
        self.logger = logging.getLogger(__name__)

    def connection_established(self) -> None:
        """Handle connection established event."""
        self.logger.debug(f"Connection established, sending {self.raw_input}")
        self.sendFrame(self.raw_input.encode())

    def telegram_sent(self, telegram_sent: str) -> None:
        """Handle telegram sent event.

        Args:
            telegram_sent: The telegram that was sent.
        """
        self.service_response.success = True
        self.service_response.sent_telegrams = telegram_sent
        self.service_response.timestamp = datetime.now()
        self.service_response.received_telegrams = []

    def telegram_received(self, telegram_received: TelegramReceivedEvent) -> None:
        """Handle telegram received event.

        Args:
            telegram_received: The telegram received event.
        """
        self.logger.debug(f"Telegram received: {telegram_received}")
        if not self.service_response.received_telegrams:
            self.service_response.received_telegrams = []
        self.service_response.received_telegrams.append(telegram_received.frame)

        if self.progress_callback:
            self.progress_callback(telegram_received.frame)

    def timeout(self) -> bool:
        """Handle timeout event.

        Returns:
            False to indicate connection should be closed.
        """
        self.logger.debug(f"Timeout: {self.timeout_seconds}s")
        if self.finish_callback:
            self.finish_callback(self.service_response)
        return False

    def failed(self, message: str) -> None:
        """Handle failed connection event.

        Args:
            message: Failure message.
        """
        self.logger.debug(f"Failed with message: {message}")
        self.service_response.success = False
        self.service_response.timestamp = datetime.now()
        self.service_response.error = message
        if self.finish_callback:
            self.finish_callback(self.service_response)

    def send_raw_telegram(
        self,
        raw_input: str,
        progress_callback: Callable[[str], None],
        finish_callback: Callable[[ConbusRawResponse], None],
        timeout_seconds: Optional[float] = None,
    ) -> None:
        """Send a raw telegram string to the Conbus server.

        Args:
            raw_input: Raw telegram string to send.
            progress_callback: Callback to handle progress updates.
            finish_callback: Callback function to call when the operation is complete.
            timeout_seconds: Timeout in seconds.
        """
        self.logger.info("Starting send_raw_telegram")
        if timeout_seconds:
            self.timeout_seconds = timeout_seconds
        self.progress_callback = progress_callback
        self.finish_callback = finish_callback
        self.raw_input = raw_input
        self.start_reactor()



"""
Unit tests for Latin-1 encoding edge cases in Conbus communication.

Tests the specific encoding fix for the issue described in doc/Fix-Encoding-Issue.md
where UTF-8 decoding fails on Latin-1 characters like 0xa7 (§ symbol).
"""

import socket
import threading
import time
from contextlib import suppress
from typing import Optional


class Latin1TestServer:
    """Test server that sends responses with Latin-1 extended characters."""

    def __init__(self, port=10003):
        """Initialize the Latin1 test server.

        Args:
            port: Port number to listen on (default: 10003).
        """
        self.port = port
        self.server_socket: Optional[socket.socket] = None
        self.is_running = False
        self.received_messages = []

    def start(self):
        """Start the test server."""
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind(("localhost", self.port))
        self.server_socket.listen(1)
        self.is_running = True

        server_thread = threading.Thread(target=self._accept_connections, daemon=True)
        server_thread.start()
        time.sleep(0.1)  # Give server time to start

    def stop(self):
        """Stop the test server."""
        self.is_running = False
        if self.server_socket:
            self.server_socket.close()

    def _accept_connections(self):
        """Accept and handle client connections."""
        while self.is_running:
            try:
                if not self.server_socket:
                    raise socket.error
                client_socket, addr = self.server_socket.accept()
                self._handle_client(client_socket)
            except (socket.error, OSError):
                break

    def _handle_client(self, client_socket):
        """Handle individual client connection."""
        try:
            client_socket.settimeout(2.0)

            while True:
                data = client_socket.recv(1024)
                if not data:
                    break

                message = data.decode("latin-1").strip()
                self.received_messages.append(message)

                # Send responses with Latin-1 extended characters
                response = self._generate_latin1_response(message)
                if response:
                    client_socket.send(response.encode("latin-1"))

        except socket.timeout:
            pass
        except (ValueError, KeyError, ConnectionError):
            pass
        finally:
            with suppress(ValueError, KeyError, ConnectionError):
                client_socket.close()

    @staticmethod
    def _generate_latin1_response(message):
        """Generate responses containing Latin-1 extended characters."""
        # Map of requests to responses with extended characters
        return {
            # Temperature request with § symbol (0xa7)
            "<S0020012521F02D18FN>": "<R0020044966F02D18+31,5§CIE>",
            # VOLTAGE request with © symbol (0xa9)
            "<S0012345011F02D20FM>": "<R0012345011F02D20+12,5V©OK>",
            # Current request with ® symbol (0xae)
            "<S0012345006F02D21FL>": "<R0012345006F02D21+2,3A®OK>",
            # Humidity request with ± symbol (0xb1)
            "<S0012345003F02D19FH>": "<R0012345003F02D19+65,2%±OK>",
            # Custom request with multiple extended chars
            "<S0012345011F02DE2CJ>": "<R0012345011F02DE2COUCOU§©®±FM>",
            # Discover with extended chars in device name
            "<S0000000000F01D00FA>": "<R0012345011F01D©XP24®>",
        }.get(message)


class TestEncodingConsistency:
    """Test encoding consistency across the communication pipeline."""

    def test_round_trip_encoding(self):
        """Test that messages can be encoded and decoded consistently."""
        test_messages = [
            "<S0020012521F02D18FN>",  # Normal ASCII message
            "<R0020012521F02D18+31,5§CIE>",  # Message with § symbol
            "<R0012345011F02D20+12,5V©OK>",  # Message with © symbol
            "<R0012345006F02D21+2,3A®OK>",  # Message with ® symbol
            "<R0012345003F02D19+65,2%±OK>",  # Message with ± symbol
        ]

        for message in test_messages:
            # Encode to bytes using Latin-1
            encoded = message.encode("latin-1")

            # Decode back using Latin-1
            decoded = encoded.decode("latin-1")

            # Should be identical
            assert decoded == message

            # Verify extended characters are preserved
            for char in message:
                if ord(char) > 127:  # Extended character
                    assert char in decoded

    def test_latin1_character_range(self):
        """Test that all Latin-1 characters (0-255) can be handled."""
        # Test all possible byte values
        for byte_value in range(256):
            char = chr(byte_value)
            test_string = f"Test{char}Message"

            # Should be able to encode and decode without errors
            encoded = test_string.encode("latin-1")
            decoded = encoded.decode("latin-1")

            assert decoded == test_string
            assert char in decoded



"""Output formatting utilities for CLI commands."""

import json
from typing import Any, Dict, Optional


class OutputFormatter:
    """Handles standardized output formatting for CLI commands."""

    def __init__(self, json_output: bool = False):
        """Initialize the output formatter.

        Args:
            json_output: Whether to format output as JSON (default: False).
        """
        self.json_output = json_output

    def success_response(self, data: Dict[str, Any]) -> str:
        """Format a successful response.

        Args:
            data: Response data to format.

        Returns:
            Formatted success response as string.
        """
        if self.json_output:
            return json.dumps(data, indent=2)
        return self._format_text_response(data)

    def error_response(
        self, error: str, extra_data: Optional[Dict[str, Any]] = None
    ) -> str:
        """Format an error response.

        Args:
            error: Error message.
            extra_data: Additional error data to include.

        Returns:
            Formatted error response as string.
        """
        error_data = {"success": False, "error": error}
        if extra_data:
            error_data.update(extra_data)

        if self.json_output:
            return json.dumps(error_data, indent=2)
        return f"Error: {error}"

    def validation_response(self, is_valid: bool, data: Dict[str, Any]) -> str:
        """Format a validation response.

        Args:
            is_valid: Whether validation passed.
            data: Validation data to include.

        Returns:
            Formatted validation response as string.
        """
        if self.json_output:
            response_data = {"valid": is_valid} | data
            return json.dumps(response_data, indent=2)

        status = "✓ Valid" if is_valid else "✗ Invalid"
        return f"Status: {status}"

    def checksum_status(self, is_valid: bool) -> str:
        """Format checksum validation status.

        Args:
            is_valid: Whether checksum is valid.

        Returns:
            Formatted checksum status as string.
        """
        if self.json_output:
            return json.dumps({"checksum_valid": is_valid}, indent=2)

        return "✓ Valid" if is_valid else "✗ Invalid"

    @staticmethod
    def _format_text_response(data: Dict[str, Any]) -> str:
        """Format data for human-readable text output.

        Args:
            data: Data dictionary to format.

        Returns:
            Formatted text output as string.
        """
        lines = []

        # Handle common data patterns
        if "telegram" in data:
            lines.append(f"Telegram: {data['telegram']}")

        if "serial_number" in data:
            lines.append(f"Serial: {data['serial_number']}")

        if "operation" in data:
            lines.append(f"Operation: {data['operation']}")

        if "count" in data:
            lines.append(f"Count: {data['count']}")

        # Add any remaining fields
        for key, value in data.items():
            if key not in ("telegram", "serial_number", "operation", "count"):
                if isinstance(value, (str, int, float)):
                    lines.append(f"{key.replace('_', ' ').title()}: {value}")

        return "\n".join(lines)


class TelegramFormatter(OutputFormatter):
    """Specialized formatter for telegram-related output."""

    def format_telegram_summary(
        self, telegram_data: Dict[str, Any], service_formatter_method: Any = None
    ) -> str:
        """Format telegram summary using service method when available.

        Args:
            telegram_data: Telegram data to format.
            service_formatter_method: Optional service formatter method.

        Returns:
            Formatted telegram summary as string.
        """
        if self.json_output:
            return json.dumps(telegram_data, indent=2)

        if service_formatter_method:
            return str(service_formatter_method)

        # Fallback formatting
        lines = []
        if "telegram_type" in telegram_data:
            lines.append(f"Type: {telegram_data['telegram_type'].title()}")
        if "raw_telegram" in telegram_data:
            lines.append(f"Raw: {telegram_data['raw_telegram']}")
        if "timestamp" in telegram_data:
            lines.append(f"Timestamp: {telegram_data['timestamp']}")

        return "\n".join(lines)

    def format_validation_result(
        self, parsed_telegram: Any, checksum_valid: Optional[bool], service_summary: str
    ) -> str:
        """Format telegram validation results.

        Args:
            parsed_telegram: Parsed telegram object.
            checksum_valid: Whether checksum is valid.
            service_summary: Summary from service.

        Returns:
            Formatted validation result as string.
        """
        if self.json_output:
            output = parsed_telegram.to_dict()
            output["checksum_valid"] = checksum_valid
            return json.dumps(output, indent=2)

        lines = [service_summary]
        if checksum_valid is not None:
            status = "✓ Valid" if checksum_valid else "✗ Invalid"
            lines.append(f"Checksum validation: {status}")

        return "\n".join(lines)


class ListFormatter(OutputFormatter):
    """Specialized formatter for list-based output."""

    def format_list_response(
        self, items: list, title: str, item_formatter: Any = None
    ) -> str:
        """Format a list of items with optional custom formatter.

        Args:
            items: List of items to format.
            title: Title for the list.
            item_formatter: Optional custom formatter function.

        Returns:
            Formatted list as string.
        """
        if self.json_output:
            return json.dumps(
                {
                    "items": [
                        item.to_dict() if hasattr(item, "to_dict") else item
                        for item in items
                    ],
                    "count": len(items),
                },
                indent=2,
            )

        lines = [f"{title}: {len(items)} items", "-" * 50]

        for i, item in enumerate(items, 1):
            if item_formatter:
                lines.append(f"{i}. {item_formatter(item)}")
            elif hasattr(item, "__str__"):
                lines.append(f"{i}. {item}")
            else:
                lines.append(f"{i}. {item}")

        return "\n".join(lines)

    def format_search_results(self, matches: list, query: str) -> str:
        """Format search results.

        Args:
            matches: List of matching items.
            query: Search query string.

        Returns:
            Formatted search results as string.
        """
        if self.json_output:
            return json.dumps(
                {
                    "query": query,
                    "matches": [
                        item.to_dict() if hasattr(item, "to_dict") else item
                        for item in matches
                    ],
                    "count": len(matches),
                },
                indent=2,
            )

        if not matches:
            return f"No items found matching '{query}'"

        lines = [f"Found {len(matches)} items matching '{query}':", "-" * 60]
        for item in matches:
            if (
                hasattr(item, "code")
                and hasattr(item, "name")
                and hasattr(item, "description")
            ):
                lines.append(f"{item.code:2} - {item.name}: {item.description}")
            else:
                lines.append(str(item))

        return "\n".join(lines)


class StatisticsFormatter(OutputFormatter):
    """Specialized formatter for statistics and analysis output."""

    def format_file_statistics(
        self, file_path: str, stats: Dict[str, Any], entry_count: int
    ) -> str:
        """Format file analysis statistics.

        Args:
            file_path: Path to the analyzed file.
            stats: Statistics dictionary.
            entry_count: Total number of entries.

        Returns:
            Formatted statistics as string.
        """
        if self.json_output:
            return json.dumps(
                {
                    "file_path": file_path,
                    "statistics": stats,
                    "entry_count": entry_count,
                },
                indent=2,
            )

        lines = [
            "=== Console Bus Log Summary ===",
            f"File: {file_path}",
            f"Entries: {entry_count}",
        ]

        # Time range
        if stats.get("time_range", {}).get("start"):
            time_range = stats["time_range"]
            lines.extend(
                [
                    f"Time Range: {time_range['start']} - {time_range['end']}",
                    f"Duration: {time_range['duration_seconds']:.3f} seconds",
                ]
            )

        # Telegram distribution
        lines.append("\nTelegram Distribution:")
        type_counts = stats.get("telegram_type_counts", {})
        total = stats.get("total_entries", 0)

        for t_type, count in type_counts.items():
            percentage = (count / total * 100) if total > 0 else 0
            lines.append(f"  {t_type.capitalize()}: {count} ({percentage:.1f}%)")

        # Direction distribution
        lines.append("\nDirection Distribution:")
        dir_counts = stats.get("direction_counts", {})
        for direction, count in dir_counts.items():
            percentage = (count / total * 100) if total > 0 else 0
            lines.append(f"  {direction.upper()}: {count} ({percentage:.1f}%)")

        return "\n".join(lines)



"""XP20 Action Table models for input actions and settings."""

from dataclasses import dataclass, field


@dataclass
class InputChannel:
    """Configuration for a single input channel in XP20 action table.

    Attributes:
        invert: Input inversion flag
        short_long: Short/long press detection flag
        group_on_off: Group on/off function flag
        and_functions: 8-bit AND function configuration array
        sa_function: SA function flag
        ta_function: TA function flag
    """

    invert: bool = False
    short_long: bool = False
    group_on_off: bool = False
    and_functions: list[bool] = field(default_factory=lambda: [False] * 8)
    sa_function: bool = False
    ta_function: bool = False


@dataclass
class Xp20MsActionTable:
    """XP20 Action Table for managing 8 input channels.

    Contains configuration for 8 input channels (input1 through input8),
    each with flags for inversion, short/long press detection, group functions,
    AND functions, SA functions, and TA functions.

    Attributes:
        input1: Configuration for input channel 1.
        input2: Configuration for input channel 2.
        input3: Configuration for input channel 3.
        input4: Configuration for input channel 4.
        input5: Configuration for input channel 5.
        input6: Configuration for input channel 6.
        input7: Configuration for input channel 7.
        input8: Configuration for input channel 8.
    """

    input1: InputChannel = field(default_factory=InputChannel)
    input2: InputChannel = field(default_factory=InputChannel)
    input3: InputChannel = field(default_factory=InputChannel)
    input4: InputChannel = field(default_factory=InputChannel)
    input5: InputChannel = field(default_factory=InputChannel)
    input6: InputChannel = field(default_factory=InputChannel)
    input7: InputChannel = field(default_factory=InputChannel)
    input8: InputChannel = field(default_factory=InputChannel)



"""Telegram Service for parsing XP telegrams.

This module provides telegram parsing functionality for event, system, and reply telegrams.
"""

import logging
import re
from typing import Union

from xp.models import EventType
from xp.models.telegram.datapoint_type import DataPointType
from xp.models.telegram.event_telegram import EventTelegram
from xp.models.telegram.output_telegram import OutputTelegram
from xp.models.telegram.reply_telegram import ReplyTelegram
from xp.models.telegram.system_function import SystemFunction
from xp.models.telegram.system_telegram import SystemTelegram
from xp.models.telegram.telegram_type import TelegramType
from xp.utils.checksum import calculate_checksum


class TelegramParsingError(Exception):
    """Raised when telegram parsing fails."""

    pass


class TelegramService:
    """Service for parsing event telegrams from the console bus.

    Handles parsing of telegrams in the format:
    <[EO]{module_type}L{link_number}I{output_number}{event_type}{checksum}>

    Attributes:
        EVENT_TELEGRAM_PATTERN: Regex pattern for event telegrams.
        SYSTEM_TELEGRAM_PATTERN: Regex pattern for system telegrams.
        REPLY_TELEGRAM_PATTERN: Regex pattern for reply telegrams.
    """

    # <O06L00I07MAG>
    # <O06L00I07BAJ>
    # <E13L12I02BAB>
    EVENT_TELEGRAM_PATTERN = re.compile(
        r"^<([EO])(\d{1,2})L(\d{2})I(\d{2})([MB])([A-Z0-9]{2})>$"
    )

    SYSTEM_TELEGRAM_PATTERN = re.compile(r"^<S(\d{10})F(\d{2})D(.{2,})([A-Z0-9]{2})>$")

    REPLY_TELEGRAM_PATTERN = re.compile(r"^<R(\d{10})F(\d{2})(.+?)([A-Z0-9]{2})>$")

    def __init__(self) -> None:
        """Initialize the telegram service."""
        # Set up logging
        self.logger = logging.getLogger(__name__)

    def parse_event_telegram(self, raw_telegram: str) -> EventTelegram:
        """Parse a raw telegram string into an EventTelegram object.

        Args:
            raw_telegram: The raw telegram string (e.g., "<E14L00I02MAK>").

        Returns:
            EventTelegram object with parsed data.

        Raises:
            TelegramParsingError: If the telegram format is invalid.
        """
        if not raw_telegram:
            raise TelegramParsingError("Empty telegram string")

        # Validate and parse using regex
        match = self.EVENT_TELEGRAM_PATTERN.match(raw_telegram.strip())
        if not match:
            raise TelegramParsingError(f"Invalid telegram format: {raw_telegram}")

        try:
            event_telegram_type = match.group(1)
            module_type = int(match.group(2))
            link_number = int(match.group(3))
            output_number = int(match.group(4))
            event_type_char = match.group(5)
            checksum = match.group(6)

            # Validate ranges
            if event_telegram_type not in ("E", "O"):
                raise TelegramParsingError(
                    f"Event telegram type (E or O): {event_telegram_type}"
                )

            if not (0 <= link_number <= 99):
                raise TelegramParsingError(
                    f"Link number out of range (0-99): {link_number}"
                )

            if not (0 <= output_number <= 99):
                raise TelegramParsingError(
                    f"Input number out of range (0-99): {output_number}"
                )

            # Parse event type
            try:
                event_type = EventType(event_type_char)
            except ValueError:
                raise TelegramParsingError(f"Invalid event type: {event_type_char}")

            # Create the telegram object
            telegram = EventTelegram(
                module_type=module_type,
                link_number=link_number,
                input_number=output_number,
                event_type=event_type,
                checksum=checksum,
                raw_telegram=raw_telegram,
            )

            # Automatically validate checksum
            telegram.checksum_validated = self.validate_checksum(telegram)

            return telegram

        except ValueError as e:
            raise TelegramParsingError(f"Invalid numeric values in telegram: {e}")

    @staticmethod
    def validate_checksum(
        telegram: Union[EventTelegram, ReplyTelegram, SystemTelegram, OutputTelegram],
    ) -> bool:
        """Validate the checksum of a parsed telegram.

        Args:
            telegram: The parsed telegram.

        Returns:
            True if checksum is valid, False otherwise.
        """
        if not telegram.checksum or len(telegram.checksum) != 2:
            return False

        # Extract the data part (everything between < and checksum)
        raw = telegram.raw_telegram
        if not raw.startswith("<") or not raw.endswith(">"):
            return False

        # Get the data part without brackets and checksum
        data_part = raw[1:-3]  # Remove '<' and last 2 chars (checksum) + '>'

        # Calculate expected checksum
        expected_checksum = calculate_checksum(data_part)

        return telegram.checksum == expected_checksum

    @staticmethod
    def format_event_telegram_summary(telegram: EventTelegram) -> str:
        """Format a telegram for human-readable output.

        Args:
            telegram: The parsed telegram.

        Returns:
            Formatted string summary.
        """
        checksum_status = ""
        if telegram.checksum_validated is not None:
            status_indicator = "✓" if telegram.checksum_validated else "✗"
            checksum_status = f" ({status_indicator})"

        return (
            f"Event: {telegram}\n"
            f"Raw: {telegram.raw_telegram}\n"
            f"Timestamp: {telegram.timestamp}\n"
            f"Checksum: {telegram.checksum}{checksum_status}"
        )

    def parse_system_telegram(self, raw_telegram: str) -> SystemTelegram:
        """Parse a raw system telegram string into a SystemTelegram object.

        Args:
            raw_telegram: The raw telegram string (e.g., "<S0020012521F02D18FN>").

        Returns:
            SystemTelegram object with parsed data.

        Raises:
            TelegramParsingError: If the telegram format is invalid.
        """
        if not raw_telegram:
            raise TelegramParsingError("Empty telegram string")

        # Validate and parse using regex
        match = self.SYSTEM_TELEGRAM_PATTERN.match(raw_telegram.strip())
        if not match:
            raise TelegramParsingError(
                f"Invalid system telegram format: {raw_telegram}"
            )

        try:
            serial_number = match.group(1)
            function_code = match.group(2)
            data = match.group(3)
            checksum = match.group(4)

            # Parse system function
            system_function = SystemFunction.from_code(function_code)
            if system_function is None:
                raise TelegramParsingError(
                    f"Unknown system function code: {function_code}"
                )

            # Parse data point type
            datapoint_type = None
            if system_function == SystemFunction.READ_DATAPOINT:
                datapoint_type = DataPointType.from_code(data)

            # Create the telegram object
            telegram = SystemTelegram(
                serial_number=serial_number,
                system_function=system_function,
                data=data,
                datapoint_type=datapoint_type,
                checksum=checksum,
                raw_telegram=raw_telegram,
            )

            # Automatically validate checksum
            telegram.checksum_validated = self.validate_checksum(telegram)

            return telegram

        except ValueError as e:
            raise TelegramParsingError(f"Invalid values in system telegram: {e}")

    def parse_reply_telegram(self, raw_telegram: str) -> ReplyTelegram:
        """Parse a raw reply telegram string into a ReplyTelegram object.

        Args:
            raw_telegram: The raw telegram string (e.g., "<R0020012521F02D18+26,0§CIL>").

        Returns:
            ReplyTelegram object with parsed data.

        Raises:
            TelegramParsingError: If the telegram format is invalid.
        """
        if not raw_telegram:
            raise TelegramParsingError("Empty telegram string")

        # Validate and parse using regex
        self.logger.debug(f"Parsing reply telegram {raw_telegram}")
        match = self.REPLY_TELEGRAM_PATTERN.match(raw_telegram.strip())
        if not match:
            raise TelegramParsingError(f"Invalid reply telegram format: {raw_telegram}")

        try:
            serial_number = match.group(1)
            function_code = match.group(2)
            full_data_value = match.group(3)
            checksum = match.group(4)

            # Parse system function
            system_function = SystemFunction.from_code(function_code)
            if system_function is None:
                raise TelegramParsingError(
                    f"Unknown system function code: {function_code}"
                )

            # Parse data point and data value from full_data_value
            if full_data_value.startswith("D") and len(full_data_value) >= 3:
                # Regular reply format: D{data_point}{data}
                data = full_data_value[1:3]
                data_value = full_data_value[3:] if len(full_data_value) > 3 else ""
            else:
                # ACK/NAK format: just data (like "D" for ACK/NAK)
                data = "00"  # Default to STATUS
                data_value = full_data_value

            # Parse data point type
            data_point_type = DataPointType.from_code(data)

            # Create the telegram object
            telegram = ReplyTelegram(
                serial_number=serial_number,
                system_function=system_function,
                data=data,
                datapoint_type=data_point_type,
                data_value=data_value,
                checksum=checksum,
                raw_telegram=raw_telegram,
            )

            # Automatically validate checksum
            telegram.checksum_validated = self.validate_checksum(telegram)

            return telegram

        except ValueError as e:
            raise TelegramParsingError(f"Invalid values in reply telegram: {e}")

    def parse_telegram(
        self, raw_telegram: str
    ) -> Union[EventTelegram, SystemTelegram, ReplyTelegram]:
        """Auto-detect and parse any type of telegram.

        Args:
            raw_telegram: The raw telegram string.

        Returns:
            Appropriate telegram object based on type.

        Raises:
            TelegramParsingError: If the telegram format is invalid or unknown.
        """
        if not raw_telegram:
            raise TelegramParsingError("Empty telegram string")

        # Then check general telegram types
        telegram_type_code = (
            raw_telegram.strip()[1] if len(raw_telegram.strip()) > 1 else ""
        )

        if telegram_type_code in (TelegramType.EVENT.value, TelegramType.CPEVENT.value):
            return self.parse_event_telegram(raw_telegram)
        elif telegram_type_code == TelegramType.SYSTEM.value:
            return self.parse_system_telegram(raw_telegram)
        elif telegram_type_code == TelegramType.REPLY.value:
            return self.parse_reply_telegram(raw_telegram)
        else:
            raise TelegramParsingError(
                f"Unknown telegram type code: {telegram_type_code}"
            )

    @staticmethod
    def format_system_telegram_summary(telegram: SystemTelegram) -> str:
        """Format a system telegram for human-readable output.

        Args:
            telegram: The parsed system telegram.

        Returns:
            Formatted string summary.
        """
        checksum_status = ""
        if telegram.checksum_validated is not None:
            status_indicator = "✓" if telegram.checksum_validated else "✗"
            checksum_status = f" ({status_indicator})"

        return (
            f"System: {telegram}\n"
            f"Raw: {telegram.raw_telegram}\n"
            f"Timestamp: {telegram.timestamp}\n"
            f"Checksum: {telegram.checksum}{checksum_status}"
        )

    @staticmethod
    def format_reply_telegram_summary(telegram: ReplyTelegram) -> str:
        """Format a reply telegram for human-readable output.

        Args:
            telegram: The parsed reply telegram.

        Returns:
            Formatted string summary.
        """
        parsed_data = telegram.parse_datapoint_value
        data_display = (
            parsed_data.get("formatted", telegram.data_value)
            if parsed_data.get("parsed")
            else telegram.data_value
        )

        checksum_status = ""
        if telegram.checksum_validated is not None:
            status_indicator = "✓" if telegram.checksum_validated else "✗"
            checksum_status = f" ({status_indicator})"

        return (
            f"Reply: {telegram}\n"
            f"Data: {data_display}\n"
            f"Raw: {telegram.raw_telegram}\n"
            f"Timestamp: {telegram.timestamp}\n"
            f"Checksum: {telegram.checksum}{checksum_status}"
        )



"""ActionTable CLI commands."""

import json
from contextlib import suppress
from pathlib import Path
from typing import Any

import click
from click import Context

from xp.cli.commands.conbus.conbus import conbus_actiontable
from xp.cli.utils.decorators import (
    connection_command,
)
from xp.cli.utils.serial_number_type import SERIAL
from xp.models.actiontable.actiontable import ActionTable
from xp.models.homekit.homekit_conson_config import (
    ConsonModuleConfig,
    ConsonModuleListConfig,
)
from xp.services.conbus.actiontable.actiontable_download_service import (
    ActionTableService,
)
from xp.services.conbus.actiontable.actiontable_list_service import (
    ActionTableListService,
)
from xp.services.conbus.actiontable.actiontable_show_service import (
    ActionTableShowService,
)
from xp.services.conbus.actiontable.actiontable_upload_service import (
    ActionTableUploadService,
)


class ActionTableError(Exception):
    """Raised when ActionTable operations fail."""

    pass


@conbus_actiontable.command("download", short_help="Download ActionTable")
@click.argument("serial_number", type=SERIAL)
@click.pass_context
@connection_command()
def conbus_download_actiontable(ctx: Context, serial_number: str) -> None:
    """Download action table from XP module.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.
    """
    service: ActionTableService = (
        ctx.obj.get("container").get_container().resolve(ActionTableService)
    )

    def progress_callback(progress: str) -> None:
        """Handle progress updates during action table download.

        Args:
            progress: Progress message string.
        """
        click.echo(progress)

    def on_finish(
        _actiontable: ActionTable,
        actiontable_dict: dict[str, Any],
        actiontable_short: list[str],
    ) -> None:
        """Handle successful completion of action table download.

        Args:
            _actiontable: Downloaded action table object.
            actiontable_dict: Dictionary representation of action table.
            actiontable_short: List of textual format strings.
        """
        output = {
            "serial_number": serial_number,
            "actiontable_short": actiontable_short,
            "actiontable": actiontable_dict,
        }
        click.echo(json.dumps(output, indent=2, default=str))

    def error_callback(error: str) -> None:
        """Handle errors during action table download.

        Args:
            error: Error message string.
        """
        click.echo(error)

    with service:
        service.start(
            serial_number=serial_number,
            progress_callback=progress_callback,
            finish_callback=on_finish,
            error_callback=error_callback,
        )


@conbus_actiontable.command("upload", short_help="Upload ActionTable")
@click.argument("serial_number", type=SERIAL)
@click.pass_context
@connection_command()
def conbus_upload_actiontable(ctx: Context, serial_number: str) -> None:
    """Upload action table from conson.yml to XP module.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.
    """
    service: ActionTableUploadService = (
        ctx.obj.get("container").get_container().resolve(ActionTableUploadService)
    )

    click.echo(f"Uploading action table to {serial_number}...")

    # Track number of entries for success message
    entries_count = 0

    def progress_callback(progress: str) -> None:
        """Handle progress updates during action table upload.

        Args:
            progress: Progress message string.
        """
        click.echo(progress, nl=False)

    def success_callback() -> None:
        """Handle successful completion of action table upload."""
        click.echo("\nAction table uploaded successfully")
        if entries_count > 0:
            click.echo(f"{entries_count} entries written")

    def error_callback(error: str) -> None:
        """Handle errors during action table upload.

        Args:
            error: Error message string.

        Raises:
            ActionTableError: Always raised with upload failure message.
        """
        raise ActionTableError(f"Upload failed: {error}")

    with service:
        # Load config to get entry count for success message
        config_path = Path.cwd() / "conson.yml"
        if config_path.exists():
            with suppress(Exception):
                config = ConsonModuleListConfig.from_yaml(str(config_path))
                module = config.find_module(serial_number)
                if module and module.action_table:
                    entries_count = len(module.action_table)

        service.start(
            serial_number=serial_number,
            progress_callback=progress_callback,
            success_callback=success_callback,
            error_callback=error_callback,
        )


@conbus_actiontable.command("list", short_help="List modules with ActionTable")
@click.pass_context
def conbus_list_actiontable(ctx: Context) -> None:
    """List all modules with action table configurations from conson.yml.

    Args:
        ctx: Click context object.
    """
    service: ActionTableListService = (
        ctx.obj.get("container").get_container().resolve(ActionTableListService)
    )

    def on_finish(module_list: dict) -> None:
        """Handle successful completion of action table list.

        Args:
            module_list: Dictionary containing modules and total count.
        """
        click.echo(json.dumps(module_list, indent=2, default=str))

    def error_callback(error: str) -> None:
        """Handle errors during action table list.

        Args:
            error: Error message string.
        """
        click.echo(error)

    with service:
        service.start(
            finish_callback=on_finish,
            error_callback=error_callback,
        )


@conbus_actiontable.command("show", short_help="Show ActionTable configuration")
@click.argument("serial_number", type=SERIAL)
@click.pass_context
def conbus_show_actiontable(ctx: Context, serial_number: str) -> None:
    """Show action table configuration for a specific module from conson.yml.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.
    """
    service: ActionTableShowService = (
        ctx.obj.get("container").get_container().resolve(ActionTableShowService)
    )

    def on_finish(module: ConsonModuleConfig) -> None:
        """Handle successful completion of action table show.

        Args:
            module: Dictionary containing module configuration.
        """
        click.echo(json.dumps(module.model_dump(), indent=2, default=str))

    def error_callback(error: str) -> None:
        """Handle errors during action table show.

        Args:
            error: Error message string.
        """
        click.echo(error)

    with service:
        service.start(
            serial_number=serial_number,
            finish_callback=on_finish,
            error_callback=error_callback,
        )



"""Checksum utility functions for protocol interoperability.

This module provides standard checksum calculation functions for protocol
communication compatibility, including XOR checksum and IEEE 802.3 CRC32.
Implemented for interoperability purposes under fair use provisions.

Copyright (c) 2025 ld
Licensed under MIT License - see LICENSE file for details.
"""

from xp.utils.serialization import nibble


def calculate_checksum(buffer: str) -> str:
    """Calculate simple XOR checksum of a string buffer.

    Args:
        buffer: Input string to calculate checksum for

    Returns:
        Two-character checksum string in nibble format
    """
    cc = 0
    for char in buffer:
        cc ^= ord(char)

    return nibble(cc & 0xFF)


def calculate_checksum32(buffer: bytes) -> str:
    """Calculate CRC32 checksum for protocol interoperability.

    Implements standard CRC32 algorithm using IEEE 802.3 polynomial 0xEDB88320
    for interoperability with XP protocol communications. This is a standard
    algorithm implementation for protocol compatibility purposes.

    Args:
        buffer: Byte array to calculate checksum for

    Returns:
        Eight-character checksum string in nibble format
    """
    nibble_result = ""
    crc = 0xFFFFFFFF  # Initialize to -1 (all bits set)

    for byte in buffer:
        crc ^= byte
        for _ in range(8):
            if crc & 1:
                crc = (crc >> 1) ^ 0xEDB88320
            else:
                crc = crc >> 1

    crc ^= 0xFFFFFFFF  # Final XOR

    # Convert to nibble format (4 bytes, little-endian)
    for _ in range(4):
        nibble_result = nibble(crc & 0xFF) + nibble_result
        crc >>= 8

    return nibble_result



"""HomeKit Light Bulb Accessory.

This module provides a light bulb accessory for HomeKit integration.
"""

import logging

from bubus import EventBus
from pyhap.accessory import Accessory
from pyhap.accessory_driver import AccessoryDriver
from pyhap.const import CATEGORY_LIGHTBULB

from xp.models.homekit.homekit_config import HomekitAccessoryConfig
from xp.models.homekit.homekit_conson_config import ConsonModuleConfig
from xp.models.protocol.conbus_protocol import (
    LightBulbGetOnEvent,
    LightBulbSetOnEvent,
)


class LightBulb(Accessory):
    """HomeKit light bulb accessory.

    Attributes:
        category: HomeKit category (CATEGORY_LIGHTBULB).
        event_bus: Event bus for inter-service communication.
        logger: Logger instance.
        identifier: Unique identifier for the accessory.
        accessory: Accessory configuration.
        module: Module configuration.
        is_on: Current on/off state.
        char_on: On characteristic.
    """

    category = CATEGORY_LIGHTBULB
    event_bus: EventBus

    def __init__(
        self,
        driver: AccessoryDriver,
        module: ConsonModuleConfig,
        accessory: HomekitAccessoryConfig,
        event_bus: EventBus,
    ):
        """Initialize the light bulb accessory.

        Args:
            driver: HAP accessory driver.
            module: Module configuration.
            accessory: Accessory configuration.
            event_bus: Event bus for inter-service communication.
        """
        super().__init__(driver, accessory.description)

        self.logger = logging.getLogger(__name__)

        identifier = f"{module.serial_number}.{accessory.output_number:02d}"
        version = accessory.id
        manufacturer = "Conson"
        model = ("XP24_lightbulb",)

        self.identifier = identifier
        self.accessory = accessory
        self.module = module
        self.event_bus = event_bus
        self.is_on = False

        self.logger.info(
            "Creating Lightbulb { serial_number : %s, output_number: %s }",
            module.serial_number,
            accessory.output_number,
        )

        serv_light = self.add_preload_service("Lightbulb")

        self.set_info_service(version, manufacturer, model, identifier)

        self.char_on = serv_light.configure_char(
            "On", getter_callback=self.get_on, setter_callback=self.set_on
        )

    def set_on(self, value: bool) -> None:
        """Set the on/off state of the light bulb.

        Args:
            value: True to turn on, False to turn off.
        """
        # Emit set event
        self.logger.debug(f"set_on {value}")
        if self.is_on != value:
            self.is_on = value
            self.event_bus.dispatch(
                LightBulbSetOnEvent(
                    serial_number=self.accessory.serial_number,
                    output_number=self.accessory.output_number,
                    module=self.module,
                    accessory=self.accessory,
                    value=value,
                )
            )

    def get_on(self) -> bool:
        """Get the on/off state of the light bulb.

        Returns:
            True if on, False if off.
        """
        # Emit event and get response
        self.logger.debug("get_on")
        self.event_bus.dispatch(
            LightBulbGetOnEvent(
                serial_number=self.accessory.serial_number,
                output_number=self.accessory.output_number,
                module=self.module,
                accessory=self.accessory,
            )
        )
        self.logger.debug(f"get_on from dispatch: {self.is_on}")

        return self.is_on



"""Tests for DeviceServiceFactory."""

import pytest

from xp.services.actiontable.msactiontable_serializer import MsActionTableSerializer
from xp.services.actiontable.msactiontable_xp20_serializer import (
    Xp20MsActionTableSerializer,
)
from xp.services.actiontable.msactiontable_xp24_serializer import (
    Xp24MsActionTableSerializer,
)
from xp.services.actiontable.msactiontable_xp33_serializer import (
    Xp33MsActionTableSerializer,
)
from xp.services.server.cp20_server_service import CP20ServerService
from xp.services.server.device_service_factory import DeviceServiceFactory
from xp.services.server.xp20_server_service import XP20ServerService
from xp.services.server.xp24_server_service import XP24ServerService
from xp.services.server.xp33_server_service import XP33ServerService
from xp.services.server.xp130_server_service import XP130ServerService
from xp.services.server.xp230_server_service import XP230ServerService


@pytest.fixture
def factory():
    """Create a device service factory with mock serializers."""
    xp20_serializer = Xp20MsActionTableSerializer()
    xp24_serializer = Xp24MsActionTableSerializer()
    xp33_serializer = Xp33MsActionTableSerializer()
    ms_serializer = MsActionTableSerializer()

    return DeviceServiceFactory(
        xp20ms_serializer=xp20_serializer,
        xp24ms_serializer=xp24_serializer,
        xp33ms_serializer=xp33_serializer,
        ms_serializer=ms_serializer,
    )


class TestDeviceServiceFactoryInit:
    """Test DeviceServiceFactory initialization."""

    def test_init(self, factory):
        """Test factory initialization with serializers."""
        assert factory.xp20ms_serializer is not None
        assert factory.xp24ms_serializer is not None
        assert factory.xp33ms_serializer is not None
        assert factory.ms_serializer is not None


class TestDeviceServiceFactoryCreateDevice:
    """Test DeviceServiceFactory.create_device()."""

    def test_create_xp20(self, factory):
        """Test creating XP20 device."""
        device = factory.create_device("XP20", "12345")

        assert isinstance(device, XP20ServerService)
        assert device.serial_number == "12345"
        assert device.device_type == "XP20"

    def test_create_xp24(self, factory):
        """Test creating XP24 device."""
        device = factory.create_device("XP24", "23456")

        assert isinstance(device, XP24ServerService)
        assert device.serial_number == "23456"
        assert device.device_type == "XP24"

    def test_create_xp33(self, factory):
        """Test creating XP33 device."""
        device = factory.create_device("XP33", "33333")

        assert isinstance(device, XP33ServerService)
        assert device.serial_number == "33333"
        assert device.variant == "XP33"

    def test_create_xp33lr(self, factory):
        """Test creating XP33LR device."""
        device = factory.create_device("XP33LR", "33334")

        assert isinstance(device, XP33ServerService)
        assert device.serial_number == "33334"
        assert device.variant == "XP33LR"

    def test_create_xp33led(self, factory):
        """Test creating XP33LED device."""
        device = factory.create_device("XP33LED", "33335")

        assert isinstance(device, XP33ServerService)
        assert device.serial_number == "33335"
        assert device.variant == "XP33LED"

    def test_create_cp20(self, factory):
        """Test creating CP20 device."""
        device = factory.create_device("CP20", "44444")

        assert isinstance(device, CP20ServerService)
        assert device.serial_number == "44444"
        assert device.device_type == "CP20"

    def test_create_xp130(self, factory):
        """Test creating XP130 device."""
        device = factory.create_device("XP130", "55555")

        assert isinstance(device, XP130ServerService)
        assert device.serial_number == "55555"
        assert device.device_type == "XP130"

    def test_create_xp230(self, factory):
        """Test creating XP230 device."""
        device = factory.create_device("XP230", "66666")

        assert isinstance(device, XP230ServerService)
        assert device.serial_number == "66666"
        assert device.device_type == "XP230"

    def test_create_unknown_device(self, factory):
        """Test creating unknown device type raises ValueError."""
        with pytest.raises(ValueError) as exc_info:
            factory.create_device("UNKNOWN", "99999")

        assert "Unknown device type 'UNKNOWN'" in str(exc_info.value)
        assert "99999" in str(exc_info.value)


class TestDeviceServiceFactorySerializerInjection:
    """Test that serializers are properly injected."""

    def test_xp20_has_serializer(self, factory):
        """Test XP20 device receives correct serializer."""
        device = factory.create_device("XP20", "12345")

        assert device.msactiontable_serializer is not None
        assert isinstance(device.msactiontable_serializer, Xp20MsActionTableSerializer)

    def test_xp24_has_serializer(self, factory):
        """Test XP24 device receives correct serializer."""
        device = factory.create_device("XP24", "23456")

        assert device.msactiontable_serializer is not None
        assert isinstance(device.msactiontable_serializer, Xp24MsActionTableSerializer)

    def test_xp33_has_serializer(self, factory):
        """Test XP33 device receives correct serializer."""
        device = factory.create_device("XP33", "33333")

        assert device.msactiontable_serializer is not None
        assert isinstance(device.msactiontable_serializer, Xp33MsActionTableSerializer)



"""XP33 Server Service for device emulation.

This service provides XP33-specific device emulation functionality,
including response generation and device configuration handling for
3-channel light dimmer modules.
"""

import socket
import threading
from typing import Dict, Optional

from xp.models import ModuleTypeCode
from xp.models.actiontable.msactiontable_xp33 import Xp33MsActionTable
from xp.models.telegram.datapoint_type import DataPointType
from xp.models.telegram.system_function import SystemFunction
from xp.models.telegram.system_telegram import SystemTelegram
from xp.services.actiontable.msactiontable_xp33_serializer import (
    Xp33MsActionTableSerializer,
)
from xp.services.server.base_server_service import BaseServerService


class XP33ServerError(Exception):
    """Raised when XP33 server operations fail."""

    pass


class XP33ServerService(BaseServerService):
    """
    XP33 device emulation service.

    Generates XP33-specific responses, handles XP33 device configuration,
    and implements XP33 telegram format for 3-channel dimmer modules.
    """

    def __init__(
        self,
        serial_number: str,
        variant: str = "XP33LR",
        msactiontable_serializer: Optional[Xp33MsActionTableSerializer] = None,
    ):
        """Initialize XP33 server service.

        Args:
            serial_number: The device serial number.
            variant: Device variant (XP33, XP33LR, or XP33LED).
            msactiontable_serializer: MsActionTable serializer (injected via DI).
        """
        super().__init__(serial_number)
        self.variant = variant  # XP33 or XP33LR or XP33LED
        self.device_type = "XP33"
        self.module_type_code = ModuleTypeCode.XP33  # XP33 module type

        # XP33 device characteristics (anonymized for interoperability testing)
        if variant == "XP33LED":
            self.firmware_version = "XP33LED_V0.00.00"
            self.ean_code = "1234567890123"  # Test EAN - not a real product code
            self.max_power = 300  # 3 x 100VA
            self.module_type_code = ModuleTypeCode.XP33LED  # XP33LR module type
        elif variant == "XP33LR":  # XP33LR
            self.firmware_version = "XP33LR_V0.00.00"
            self.ean_code = "1234567890124"  # Test EAN - not a real product code
            self.max_power = 640  # Total 640VA
            self.module_type_code = ModuleTypeCode.XP33LR  # XP33LR module type
        else:  # XP33
            self.firmware_version = "XP33_V0.04.02"
            self.ean_code = "1234567890125"  # Test EAN - not a real product code
            self.max_power = 100  # Total 640VA
            self.module_type_code = ModuleTypeCode.XP33  # XP33 module type

        self.device_status = "00"  # Normal status
        self.link_number = 4  # 4 links configured
        self.autoreport_status = True

        # Channel states (3 channels, 0-100% dimming)
        self.channel_states = [0, 0, 0]  # All channels at 0%

        # Scene configuration (4 scenes)
        self.scenes = {
            1: [50, 30, 20],  # Scene 1: 50%, 30%, 20%
            2: [100, 100, 100],  # Scene 2: All full
            3: [25, 25, 25],  # Scene 3: Low level
            4: [0, 0, 0],  # Scene 4: Off
        }

        # Storm mode state (XP33 Storm Simulator)
        self.storm_mode = False  # Track if device is in storm mode
        self.last_response: Optional[str] = None  # Cache last response for storm replay
        self.storm_thread: Optional[threading.Thread] = (
            None  # Background thread for storm
        )
        self.storm_stop_event = threading.Event()  # Event to stop storm thread
        self.client_sockets: set[socket.socket] = set()  # All active client sockets
        self.client_sockets_lock = threading.Lock()  # Lock for socket set
        self.storm_packets_sent = 0  # Counter for packets sent during storm

        # MsActionTable support
        self.msactiontable_serializer = (
            msactiontable_serializer or Xp33MsActionTableSerializer()
        )
        self.msactiontable = self._get_default_msactiontable()

    def _handle_device_specific_action_request(
        self, request: SystemTelegram
    ) -> Optional[str]:
        """Handle XP33-specific action requests."""
        telegrams = self._handle_action_channel_dimming(request.data)
        self.logger.debug(f"Generated {self.device_type} action responses: {telegrams}")
        return telegrams

    def _handle_action_channel_dimming(self, data_value: str) -> str:
        """Handle XP33-specific channel dimming action.

        Args:
            data_value: Action data in format channel_number:dimming_level.
                       E.g., "00:050" means channel 0, 50% dimming.

        Returns:
            Response telegram(s) - ACK/NAK, optionally with event telegram.
        """
        if ":" not in data_value or len(data_value) < 6:
            return self._build_ack_nak_response_telegram(False)

        try:
            parts = data_value.split(":")
            channel_number = int(parts[0])
            dimming_level = int(parts[1])
        except (ValueError, IndexError):
            return self._build_ack_nak_response_telegram(False)

        if channel_number not in range(len(self.channel_states)):
            return self._build_ack_nak_response_telegram(False)

        if dimming_level not in range(0, 101):
            return self._build_ack_nak_response_telegram(False)

        previous_level = self.channel_states[channel_number]
        self.channel_states[channel_number] = dimming_level
        state_changed = (previous_level == 0 and dimming_level > 0) or (
            previous_level > 0 and dimming_level == 0
        )

        telegrams = self._build_ack_nak_response_telegram(True)
        if state_changed and self.autoreport_status:
            # Report dimming change event
            telegrams += self._build_dimming_event_telegram(
                dimming_level, channel_number
            )

        return telegrams

    def _build_ack_nak_response_telegram(self, ack_or_nak: bool) -> str:
        """Build a complete ACK or NAK response telegram with checksum.

        Args:
            ack_or_nak: true: ACK telegram response, false: NAK telegram response.

        Returns:
            The complete telegram with checksum enclosed in angle brackets.
        """
        data_value = (
            SystemFunction.ACK.value if ack_or_nak else SystemFunction.NAK.value
        )
        data_part = f"R{self.serial_number}" f"F{data_value:02}D"
        return self._build_response_telegram(data_part)

    def _build_dimming_event_telegram(
        self, dimming_level: int, channel_number: int
    ) -> str:
        """Build a complete dimming event telegram with checksum.

        Args:
            dimming_level: Dimming level 0-100%.
            channel_number: Channel concerned (0-2).

        Returns:
            The complete event telegram with checksum enclosed in angle brackets.
        """
        data_value = "M" if dimming_level > 0 else "B"
        data_part = (
            f"E{self.module_type_code.value:02}"
            f"L{self.link_number:02}"
            f"I{channel_number:02}"
            f"{data_value}"
        )
        return self._build_response_telegram(data_part)

    def _handle_device_specific_data_request(
        self, request: SystemTelegram
    ) -> Optional[str]:
        """Handle XP33-specific data requests with storm mode support."""
        if not request.datapoint_type:
            # Check for D99 storm trigger (not in DataPointType enum)
            if request.data and request.data.startswith("99"):
                return self._trigger_storm_mode()
            return None

        datapoint_type = request.datapoint_type

        # Storm mode handling
        if datapoint_type == DataPointType.MODULE_ERROR_CODE:
            if self.storm_mode:
                # MODULE_ERROR_CODE query stops storm
                return self._exit_storm_mode()
            else:
                # Normal operation - return error code 00
                return self._build_error_code_response("00")

        # If in storm mode and not MODULE_ERROR_CODE query, ignore (background thread is sending)
        if self.storm_mode:
            self.logger.debug(
                f"Ignoring query during storm mode for device {self.serial_number}"
            )
            return None  # Background thread is sending storm telegrams

        # Normal data request handling
        handler = {
            DataPointType.MODULE_OUTPUT_STATE: self._handle_read_module_output_state,
            DataPointType.MODULE_STATE: self._handle_read_module_state,
            DataPointType.MODULE_OPERATING_HOURS: self._handle_read_module_operating_hours,
            DataPointType.MODULE_LIGHT_LEVEL: self._handle_read_light_level,
        }.get(datapoint_type)
        if not handler:
            return None

        data_value = handler()
        data_part = (
            f"R{self.serial_number}"
            f"F02D{datapoint_type.value}"
            f"{self.module_type_code.value:02}"
            f"{data_value}"
        )
        telegram = self._build_response_telegram(data_part)

        # Cache response for potential storm replay
        self.last_response = telegram

        self.logger.debug(
            f"Generated {self.device_type} module type response: {telegram}"
        )
        return telegram

    def _handle_read_module_output_state(self) -> str:
        """Handle XP33-specific module output state.

        Returns:
            String representation of the output state for 3 channels.
        """
        return (
            f"xxxxx"
            f"{1 if self.channel_states[0] > 0 else 0}"
            f"{1 if self.channel_states[1] > 0 else 0}"
            f"{1 if self.channel_states[2] > 0 else 0}"
        )

    def _handle_read_module_state(self) -> str:
        """Handle XP33-specific module state.

        Returns:
            'ON' if any channel is active, 'OFF' otherwise.
        """
        if any(level > 0 for level in self.channel_states):
            return "ON"
        return "OFF"

    def _handle_read_module_operating_hours(self) -> str:
        """Handle XP33-specific module operating hours.

        Returns:
            Operating hours for all 3 channels.
        """
        return "00:000[H],01:000[H],02:000[H]"

    def _handle_read_light_level(self) -> str:
        """Handle XP33-specific light level reading.

        Returns:
            Light levels for all channels in format "00:000[%],01:000[%],02:000[%]".
        """
        levels = [
            f"{i:02d}:{level:03d}[%]" for i, level in enumerate(self.channel_states)
        ]
        return ",".join(levels)

    def _trigger_storm_mode(self) -> Optional[str]:
        """Trigger storm mode via D99 query.

        Starts a background thread that sends 2 packets per second.
        If storm is already active, this is a no-op.

        Returns:
            None (no response - storm mode activated).
        """
        # If storm already active, just log and continue
        if self.storm_mode and self.storm_thread and self.storm_thread.is_alive():
            self.logger.debug(
                f"Storm already active for device {self.serial_number}, "
                f"sent {self.storm_packets_sent}/200 packets"
            )
            return None

        if not self.last_response:
            self.logger.warning(
                f"Cannot trigger storm for device {self.serial_number}: "
                f"no cached response"
            )
            return None

        self.storm_mode = True
        self.storm_packets_sent = 0
        self.storm_stop_event.clear()

        # Start background thread to send storm telegrams
        self.storm_thread = threading.Thread(
            target=self._storm_sender_thread,
            daemon=True,
            name=f"Storm-{self.serial_number}",
        )
        self.storm_thread.start()

        self.logger.info(
            f"Storm triggered via D99 query for device {self.serial_number}"
        )
        return None  # No response when entering storm mode

    def _exit_storm_mode(self) -> str:
        """Exit storm mode and return error code FE.

        Stops the background storm thread and returns error code.

        Returns:
            MODULE_ERROR_CODE response with error code FE (buffer overflow).
        """
        self.logger.info(
            f"MODULE_ERROR_CODE query received, stopping storm for device {self.serial_number}"
        )

        # Signal the storm thread to stop
        self.storm_stop_event.set()
        self.storm_mode = False

        # Wait for thread to finish (with timeout)
        if self.storm_thread and self.storm_thread.is_alive():
            self.storm_thread.join(timeout=1.0)

        self.logger.info(
            f"Storm stopped after {self.storm_packets_sent} packets for device {self.serial_number}"
        )
        self.logger.info(
            f"Storm stopped, returning to normal operation for device {self.serial_number}"
        )
        return self._build_error_code_response("FE")

    def _storm_sender_thread(self) -> None:
        """Background thread that sends storm telegrams continuously.

        Sends 2 packets per second (500ms delay) until:
        - 200 packets have been sent, or
        - Storm mode is stopped via stop event

        The storm persists across socket disconnections. If the client disconnects
        and reconnects, the storm will continue on the new connection.
        """
        if not self.last_response:
            self.logger.error(
                f"Storm thread started but missing cached response for {self.serial_number}"
            )
            self.storm_mode = False
            return

        self.logger.info(
            f"Storm thread started, sending 200 duplicate telegrams at 2 packets/sec for device {self.serial_number}"
        )

        # Type narrowing for mypy
        cached_response: str = self.last_response
        max_packets = 200
        packets_per_second = 2
        delay_between_packets = 1.0 / packets_per_second  # 0.5 seconds

        try:
            while (
                self.storm_packets_sent < max_packets
                and not self.storm_stop_event.is_set()
            ):
                # Wait for a valid socket (client may have disconnected and reconnected)
                self.add_telegram_buffer(cached_response)
                self.storm_packets_sent += 1
                self.logger.debug(
                    f"Storm packet {self.storm_packets_sent}/{max_packets} sent for {self.serial_number}"
                )

                # Wait before sending next packet (0.5 seconds for 2 packets/sec)
                if self.storm_packets_sent < max_packets:
                    self.storm_stop_event.wait(timeout=delay_between_packets)

                # Log completion status
            if self.storm_packets_sent >= max_packets:
                self.logger.info(
                    f"Storm completed: sent all {self.storm_packets_sent} packets for {self.serial_number}"
                )
            elif self.storm_stop_event.is_set():
                self.logger.info(
                    f"Storm stopped by error code query: sent {self.storm_packets_sent} packets for {self.serial_number}"
                )

            # Clean up storm mode
            self.storm_mode = False

        except Exception as e:
            self.logger.error(
                f"Unexpected error in storm thread for {self.serial_number}: {e}"
            )
            self.storm_mode = False

    def _build_error_code_response(self, error_code: str) -> str:
        """Build MODULE_ERROR_CODE response telegram.

        Args:
            error_code: Error code (00 = normal, FE = buffer overflow).

        Returns:
            The complete MODULE_ERROR_CODE response telegram.
        """
        data_part = (
            f"R{self.serial_number}"
            f"F02D{DataPointType.MODULE_ERROR_CODE.value}"
            f"{error_code}"
        )
        telegram = self._build_response_telegram(data_part)
        self.logger.debug(
            f"Generated {self.device_type} error code response: {telegram}"
        )
        return telegram

    def set_channel_dimming(self, channel: int, level: int) -> bool:
        """Set individual channel dimming level.

        Args:
            channel: Channel number (1-3).
            level: Dimming level (0-100 percent).

        Returns:
            True if channel was set successfully, False otherwise.
        """
        if 1 <= channel <= 3 and 0 <= level <= 100:
            self.channel_states[channel - 1] = level
            self.logger.info(f"XP33 channel {channel} set to {level}%")
            return True
        return False

    def activate_scene(self, scene: int) -> bool:
        """Activate a pre-programmed scene.

        Args:
            scene: Scene number (1-4).

        Returns:
            True if scene was activated successfully, False otherwise.
        """
        if scene in self.scenes:
            self.channel_states = self.scenes[scene].copy()
            self.logger.info(f"XP33 scene {scene} activated: {self.channel_states}")
            return True
        return False

    def _get_msactiontable_serializer(self) -> Optional[Xp33MsActionTableSerializer]:
        """Get the MsActionTable serializer for XP33.

        Returns:
            The XP33 MsActionTable serializer instance.
        """
        return self.msactiontable_serializer

    def _get_msactiontable(self) -> Optional[Xp33MsActionTable]:
        """Get the MsActionTable for XP33.

        Returns:
            The XP33 MsActionTable instance.
        """
        return self.msactiontable

    def _get_default_msactiontable(self) -> Xp33MsActionTable:
        """Generate default MsActionTable configuration.

        Returns:
            Default XP33 MsActionTable with all outputs at 0-100% range, no scenes configured.
        """
        # All outputs at 0-100% range, no scenes configured
        return Xp33MsActionTable()

    def get_device_info(self) -> Dict:
        """Get XP33 device information.

        Returns:
            Dictionary containing device information.
        """
        return {
            "serial_number": self.serial_number,
            "device_type": self.device_type,
            "variant": self.variant,
            "firmware_version": self.firmware_version,
            "ean_code": self.ean_code,
            "max_power": self.max_power,
            "status": self.device_status,
            "link_number": self.link_number,
            "autoreport_status": self.autoreport_status,
            "channel_states": self.channel_states.copy(),
            "available_scenes": list(self.scenes.keys()),
        }

    def get_technical_specs(self) -> Dict:
        """Get technical specifications.

        Returns:
            Dictionary containing technical specifications.
        """
        if self.variant == "XP33LED":
            return {
                "power_per_channel": "100VA",
                "total_power": "300VA",
                "load_types": ["LED lamps", "resistive", "capacitive"],
                "dimming_type": "Leading/Trailing edge configurable",
                "protection": "Short-circuit proof channels",
            }

        # XP33LR
        return {
            "power_per_channel": "500VA max",
            "total_power": "640VA",
            "load_types": ["Resistive", "inductive"],
            "dimming_type": "Leading edge, logarithmic control",
            "protection": "Thermal protection, neutral break detection",
        }



"""Unit tests for ConbusBlinkService."""

from unittest.mock import Mock

import pytest

from xp.models.conbus.conbus_client_config import ClientConfig, ConbusClientConfig
from xp.models.telegram.system_function import SystemFunction
from xp.services.conbus.conbus_blink_service import ConbusBlinkService


class TestConbusBlinkService:
    """Unit tests for ConbusBlinkService functionality."""

    @pytest.fixture
    def mock_cli_config(self):
        """Create a test config."""
        client_config = ClientConfig(ip="10.0.0.1", port=8080, timeout=15)
        return ConbusClientConfig(conbus=client_config)

    @pytest.fixture
    def mock_reactor(self):
        """Create a mock reactor."""
        return Mock()

    @pytest.fixture
    def mock_telegram_service(self):
        """Create a mock telegram service."""
        return Mock()

    @pytest.fixture
    def service(self, mock_telegram_service, mock_cli_config, mock_reactor):
        """Create service instance with test config."""
        return ConbusBlinkService(
            telegram_service=mock_telegram_service,
            cli_config=mock_cli_config,
            reactor=mock_reactor,
        )

    def test_service_initialization(self, service):
        """Test service can be initialized with required dependencies."""
        assert service.serial_number == ""
        assert service.on_or_off == "none"
        assert service.finish_callback is None
        assert service.service_response.success is False
        assert service.service_response.system_function == SystemFunction.NONE
        assert service.service_response.operation == "none"

    def test_service_context_manager(self, service):
        """Test service can be used as context manager."""
        with service as s:
            assert s is service

    def test_connection_established_blink_on(self, service):
        """Test connection_established configures for 'on' operation."""
        from unittest.mock import patch

        service.serial_number = "0012345008"
        service.on_or_off = "on"

        # Mock send_telegram to avoid transport issues in unit tests
        with patch.object(service, "send_telegram"):
            service.connection_established()

        assert service.service_response.system_function == SystemFunction.BLINK
        assert service.service_response.operation == "on"

    def test_connection_established_blink_off(self, service):
        """Test connection_established configures for 'off' operation."""
        from unittest.mock import patch

        service.serial_number = "0012345008"
        service.on_or_off = "off"

        # Mock send_telegram to avoid transport issues in unit tests
        with patch.object(service, "send_telegram"):
            service.connection_established()

        assert service.service_response.system_function == SystemFunction.UNBLINK
        assert service.service_response.operation == "off"

    def test_telegram_sent(self, service, mock_telegram_service):
        """Test telegram_sent callback updates service response."""
        from xp.models.telegram.system_telegram import SystemTelegram

        telegram = "<S0012345008F05D00FN>"
        mock_system_telegram = SystemTelegram(
            serial_number="0012345008",
            system_function=SystemFunction.BLINK,
            raw_telegram=telegram,
            checksum="FN",
        )
        mock_telegram_service.parse_system_telegram.return_value = mock_system_telegram

        service.telegram_sent(telegram)

        assert service.service_response.sent_telegram == mock_system_telegram
        mock_telegram_service.parse_system_telegram.assert_called_once_with(telegram)

    def test_telegram_received_ack(self, service, mock_telegram_service):
        """Test telegram_received callback with ACK response."""
        from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
        from xp.models.telegram.reply_telegram import ReplyTelegram

        service.serial_number = "0012345008"

        # Mock reply telegram
        mock_reply = ReplyTelegram(
            serial_number="0012345008",
            system_function=SystemFunction.ACK,
            raw_telegram="<R0012345008F18DFA>",
            checksum="FA",
        )
        mock_telegram_service.parse_reply_telegram.return_value = mock_reply

        telegram_event = TelegramReceivedEvent(
            protocol=service,
            frame="<R0012345008F18DFA>",
            telegram="R0012345008F18DFA",
            payload="R0012345008F18D",
            telegram_type="R",
            serial_number="0012345008",
            checksum="FA",
            checksum_valid=True,
        )

        service.telegram_received(telegram_event)

        assert service.service_response.success is True
        assert service.service_response.received_telegrams == ["<R0012345008F18DFA>"]
        assert service.service_response.reply_telegram == mock_reply

    def test_telegram_received_nak(self, service, mock_telegram_service):
        """Test telegram_received callback with NAK response."""
        from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
        from xp.models.telegram.reply_telegram import ReplyTelegram

        service.serial_number = "0012345008"

        # Mock reply telegram
        mock_reply = ReplyTelegram(
            serial_number="0012345008",
            system_function=SystemFunction.NAK,
            raw_telegram="<R0012345008F19DFB>",
            checksum="FB",
        )
        mock_telegram_service.parse_reply_telegram.return_value = mock_reply

        telegram_event = TelegramReceivedEvent(
            protocol=service,
            frame="<R0012345008F19DFB>",
            telegram="R0012345008F19DFB",
            payload="R0012345008F19D",
            telegram_type="R",
            serial_number="0012345008",
            checksum="FB",
            checksum_valid=True,
        )

        service.telegram_received(telegram_event)

        assert service.service_response.success is True
        assert service.service_response.received_telegrams == ["<R0012345008F19DFB>"]
        assert service.service_response.reply_telegram == mock_reply

    def test_telegram_received_wrong_serial(self, service, mock_telegram_service):
        """Test telegram_received ignores telegrams from different serial."""
        from xp.models.protocol.conbus_protocol import TelegramReceivedEvent

        service.serial_number = "0012345008"

        telegram_event = TelegramReceivedEvent(
            protocol=service,
            frame="<R0012345999F18DFA>",
            telegram="R0012345999F18DFA",
            payload="R0012345999F18D",
            telegram_type="R",
            serial_number="0012345999",
            checksum="FA",
            checksum_valid=True,
        )

        service.telegram_received(telegram_event)

        # Should still record the telegram but not process it
        assert service.service_response.received_telegrams == ["<R0012345999F18DFA>"]
        assert service.service_response.success is False

    def test_telegram_received_with_callback(self, service, mock_telegram_service):
        """Test telegram_received calls finish callback."""
        from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
        from xp.models.telegram.reply_telegram import ReplyTelegram

        finish_mock = Mock()
        service.finish_callback = finish_mock
        service.serial_number = "0012345008"

        # Mock reply telegram
        mock_reply = ReplyTelegram(
            serial_number="0012345008",
            system_function=SystemFunction.ACK,
            raw_telegram="<R0012345008F18DFA>",
            checksum="FA",
        )
        mock_telegram_service.parse_reply_telegram.return_value = mock_reply

        telegram_event = TelegramReceivedEvent(
            protocol=service,
            frame="<R0012345008F18DFA>",
            telegram="R0012345008F18DFA",
            payload="R0012345008F18D",
            telegram_type="R",
            serial_number="0012345008",
            checksum="FA",
            checksum_valid=True,
        )

        service.telegram_received(telegram_event)

        finish_mock.assert_called_once_with(service.service_response)

    def test_failed(self, service):
        """Test failed callback updates service response."""
        finish_mock = Mock()
        service.finish_callback = finish_mock

        service.failed("Connection timeout")

        assert service.service_response.success is False
        assert service.service_response.error == "Connection timeout"
        finish_mock.assert_called_once_with(service.service_response)



"""Unit tests for ConbusEventRawService."""

from unittest.mock import Mock

import pytest

from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.services.conbus.conbus_event_raw_service import ConbusEventRawService
from xp.services.protocol.conbus_event_protocol import ConbusEventProtocol


class TestConbusEventRawService:
    """Unit tests for ConbusEventRawService functionality."""

    @pytest.fixture
    def mock_protocol(self):
        """Create a mock ConbusEventProtocol."""
        protocol = Mock(spec=ConbusEventProtocol)
        protocol.on_connection_made = Mock()
        protocol.on_telegram_sent = Mock()
        protocol.on_telegram_received = Mock()
        protocol.on_timeout = Mock()
        protocol.on_failed = Mock()
        protocol.on_connection_made.connect = Mock()
        protocol.on_telegram_sent.connect = Mock()
        protocol.on_telegram_received.connect = Mock()
        protocol.on_timeout.connect = Mock()
        protocol.on_failed.connect = Mock()
        protocol._reactor = Mock()
        protocol.telegram_queue = Mock()
        protocol.call_later = Mock()
        protocol.timeout_seconds = 5
        protocol.stop_reactor = Mock()
        return protocol

    @pytest.fixture
    def service(self, mock_protocol):
        """Create service instance with mock protocol."""
        return ConbusEventRawService(conbus_protocol=mock_protocol)

    def test_service_initialization(self, service):
        """Test service can be initialized with required dependencies."""
        assert service.module_type_code == 0
        assert service.link_number == 0
        assert service.input_number == 0
        assert service.time_ms == 1000
        assert service.finish_callback is None
        assert service.progress_callback is None
        assert service.event_result.success is False

    def test_connection_made_sends_make_event(self, service, mock_protocol):
        """Test connection_made sends MAKE event telegram."""
        service.module_type_code = 2  # CP20
        service.link_number = 10
        service.input_number = 5
        service.time_ms = 1000

        service.connection_made()

        # Verify MAKE event was queued
        mock_protocol.telegram_queue.put_nowait.assert_called()
        call_args = mock_protocol.telegram_queue.put_nowait.call_args[0][0]
        assert call_args == b"E02L10I05M"

        # Verify BREAK event was scheduled
        mock_protocol.call_later.assert_called()

    def test_telegram_sent(self, service):
        """Test telegram_sent callback updates service response."""
        telegram = "<E02L10I05MAK>"

        service.telegram_sent(telegram)

        assert service.event_result.sent_telegrams == [telegram]

    def test_telegram_received(self, service, mock_protocol):
        """Test telegram_received callback updates service response."""
        telegram_event = TelegramReceivedEvent(
            protocol=mock_protocol,
            frame="<E02L10I05MAK>",
            telegram="E02L10I05MAK",
            payload="E02L10I05M",
            telegram_type="E",
            serial_number="",
            checksum="AK",
            checksum_valid=True,
        )

        service.telegram_received(telegram_event)

        assert service.event_result.received_telegrams == ["<E02L10I05MAK>"]

    def test_telegram_received_with_progress_callback(self, service, mock_protocol):
        """Test telegram_received calls progress callback."""
        progress_mock = Mock()
        service.progress_callback = progress_mock

        telegram_event = TelegramReceivedEvent(
            protocol=mock_protocol,
            frame="<E02L10I05MAK>",
            telegram="E02L10I05MAK",
            payload="E02L10I05M",
            telegram_type="E",
            serial_number="",
            checksum="AK",
            checksum_valid=True,
        )

        service.telegram_received(telegram_event)

        progress_mock.assert_called_once_with("<E02L10I05MAK>")

    def test_timeout(self, service, mock_protocol):
        """Test timeout callback marks operation as successful."""
        finish_mock = Mock()
        service.finish_callback = finish_mock

        service.timeout()

        assert service.event_result.success is True
        assert service.event_result.error is None
        finish_mock.assert_called_once_with(service.event_result)
        mock_protocol.stop_reactor.assert_called_once()

    def test_failed(self, service, mock_protocol):
        """Test failed callback updates service response."""
        finish_mock = Mock()
        service.finish_callback = finish_mock

        service.failed("Connection timeout")

        assert service.event_result.success is False
        assert service.event_result.error == "Connection timeout"
        finish_mock.assert_called_once_with(service.event_result)
        mock_protocol.stop_reactor.assert_called_once()

    def test_run(self, service, mock_protocol):
        """Test run method sets up service parameters."""
        finish_mock = Mock()
        progress_mock = Mock()

        service.run(
            module_type_code=33,  # XP33
            link_number=15,
            input_number=8,
            time_ms=500,
            progress_callback=progress_mock,
            finish_callback=finish_mock,
            timeout_seconds=10,
        )

        assert service.module_type_code == 33
        assert service.link_number == 15
        assert service.input_number == 8
        assert service.time_ms == 500
        assert service.progress_callback == progress_mock
        assert service.finish_callback == finish_mock
        assert mock_protocol.timeout_seconds == 10

    def test_send_break_event(self, service, mock_protocol):
        """Test _send_break_event sends BREAK event telegram."""
        service.module_type_code = 2  # CP20
        service.link_number = 10
        service.input_number = 5

        service._send_break_event()

        # Verify BREAK event was queued
        mock_protocol.telegram_queue.put_nowait.assert_called()
        call_args = mock_protocol.telegram_queue.put_nowait.call_args[0][0]
        assert call_args == b"E02L10I05B"



"""Status Footer Widget for displaying app footer with connection status."""

from typing import Any, Optional

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Footer, Static

from xp.models.term.connection_state import ConnectionState
from xp.services.term.protocol_monitor_service import ProtocolMonitorService


class StatusFooterWidget(Horizontal):
    """Footer widget with connection status indicator.

    Combines the Textual Footer with a status indicator dot that shows
    the current connection state. Subscribes directly to service signals.

    Attributes:
        service: ProtocolMonitorService for connection state and status updates.
        status_widget: Static widget displaying colored status dot.
        status_text_widget: Static widget displaying status messages.
    """

    def __init__(
        self,
        service: Optional[ProtocolMonitorService] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize the Status Footer widget.

        Args:
            service: Optional ProtocolMonitorService for signal subscriptions.
            args: Additional positional arguments for Horizontal.
            kwargs: Additional keyword arguments for Horizontal.
        """
        super().__init__(*args, **kwargs)
        self.service = service
        self.status_text_widget: Static = Static("", id="status-text")
        self.status_widget: Static = Static("○", id="status-line")

    def compose(self) -> ComposeResult:
        """Compose the footer layout.

        Yields:
            Footer and status indicator widgets.
        """
        yield Footer()
        yield self.status_text_widget
        yield self.status_widget

    def on_mount(self) -> None:
        """Subscribe to service signals when widget mounts."""
        if self.service:
            self.service.on_connection_state_changed.connect(self.update_status)
            self.service.on_status_message.connect(self.update_message)

    def on_unmount(self) -> None:
        """Unsubscribe from service signals when widget unmounts."""
        if self.service:
            self.service.on_connection_state_changed.disconnect(self.update_status)
            self.service.on_status_message.disconnect(self.update_message)

    def update_status(self, state: ConnectionState) -> None:
        """Update status indicator with connection state.

        Args:
            state: Current connection state (ConnectionState enum).
        """
        # Map states to colored dots
        dot = {
            "CONNECTED": "[green]●[/green]",
            "CONNECTING": "[yellow]●[/yellow]",
            "DISCONNECTING": "[yellow]●[/yellow]",
            "FAILED": "[red]●[/red]",
            "DISCONNECTED": "○",
        }.get(state.value, "○")
        self.status_widget.update(dot)

    def update_message(self, message: str) -> None:
        """Update status text with message.

        Args:
            message: Status message to display.
        """
        self.status_text_widget.update(message)



"""Input type enumeration based on input number ranges."""

from enum import Enum


class InputType(Enum):
    """Input types based on input number ranges.

    Attributes:
        PUSH_BUTTON: Push button input (range 00-09).
        IR_REMOTE: IR remote input (range 10-89).
        PROXIMITY_SENSOR: Proximity sensor input (input 90).
    """

    PUSH_BUTTON = "push_button"  # Input 00-09
    IR_REMOTE = "ir_remote"  # Input 10-89
    PROXIMITY_SENSOR = "proximity_sensor"  # Input 90



"""Conbus client operations CLI commands."""

import json

import click
from click import Context

from xp.cli.commands.conbus.conbus import conbus
from xp.cli.utils.decorators import connection_command
from xp.cli.utils.serial_number_type import SERIAL
from xp.models import ConbusResponse
from xp.services.conbus.conbus_scan_service import ConbusScanService


@conbus.command("scan")
@click.argument("serial_number", type=SERIAL)
@click.argument("function_code", type=str)
@click.pass_context
@connection_command()
def scan_module(ctx: Context, serial_number: str, function_code: str) -> None:
    r"""Scan all datapoints of a function_code for a module.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.
        function_code: Function code.

    Examples:
        \b
        xp conbus scan 0012345011 02 # Scan all datapoints of function Read data points (02)
    """
    service: ConbusScanService = (
        ctx.obj.get("container").get_container().resolve(ConbusScanService)
    )

    def on_progress(progress: str) -> None:
        """Handle progress updates during module scan.

        Args:
            progress: Progress message string.
        """
        click.echo(progress)

    def on_finish(service_response: ConbusResponse) -> None:
        """Handle successful completion of module scan.

        Args:
            service_response: Scan response object.
        """
        click.echo(json.dumps(service_response.to_dict(), indent=2))

    with service:
        service.scan_module(
            serial_number=serial_number,
            function_code=function_code,
            progress_callback=on_progress,
            finish_callback=on_finish,
        )



"""XP33 Action Table models for output and scene configuration."""

from dataclasses import dataclass, field

from xp.models.telegram.timeparam_type import TimeParam


@dataclass
class Xp33Output:
    """Represents an XP33 output configuration.

    Attributes:
        min_level: Minimum output level (0-100).
        max_level: Maximum output level (0-100).
        scene_outputs: Enable scene outputs.
        start_at_full: Start at full brightness.
        leading_edge: Use leading edge dimming.
    """

    min_level: int = 0
    max_level: int = 100
    scene_outputs: bool = False
    start_at_full: bool = False
    leading_edge: bool = False


@dataclass
class Xp33Scene:
    """Represents a scene configuration.

    Attributes:
        output1_level: Output level for output 1 (0-100).
        output2_level: Output level for output 2 (0-100).
        output3_level: Output level for output 3 (0-100).
        time: Time parameter for scene transition.
    """

    output1_level: int = 0
    output2_level: int = 0
    output3_level: int = 0
    time: TimeParam = TimeParam.NONE


@dataclass
class Xp33MsActionTable:
    """XP33 Action Table for managing outputs and scenes.

    Attributes:
        output1: Configuration for output 1.
        output2: Configuration for output 2.
        output3: Configuration for output 3.
        scene1: Configuration for scene 1.
        scene2: Configuration for scene 2.
        scene3: Configuration for scene 3.
        scene4: Configuration for scene 4.
    """

    output1: Xp33Output = field(default_factory=Xp33Output)
    output2: Xp33Output = field(default_factory=Xp33Output)
    output3: Xp33Output = field(default_factory=Xp33Output)

    scene1: Xp33Scene = field(default_factory=Xp33Scene)
    scene2: Xp33Scene = field(default_factory=Xp33Scene)
    scene3: Xp33Scene = field(default_factory=Xp33Scene)
    scene4: Xp33Scene = field(default_factory=Xp33Scene)



"""Serializer for XP20 Action Table telegram encoding/decoding."""

from xp.models.actiontable.msactiontable_xp20 import InputChannel, Xp20MsActionTable
from xp.utils.serialization import byte_to_bits, de_nibbles, nibbles

# Index constants for clarity in implementation
SHORT_LONG_INDEX: int = 0
GROUP_ON_OFF_INDEX: int = 1
INVERT_INDEX: int = 2
AND_FUNCTIONS_INDEX: int = 3  # starts at 3, uses indices 3-10
SA_FUNCTION_INDEX: int = 11
TA_FUNCTION_INDEX: int = 12


class Xp20MsActionTableSerializer:
    """Handles serialization/deserialization of XP20 action tables to/from telegrams."""

    @staticmethod
    def to_data(action_table: Xp20MsActionTable) -> str:
        """Serialize XP20 action table to telegram hex string format.

        Args:
            action_table: XP20 action table to serialize

        Returns:
            64-character hex string (32 bytes) with A-P nibble encoding
        """
        # Initialize 32-byte raw data array
        raw_bytes = bytearray(32)

        # Get all input channels
        input_channels = [
            action_table.input1,
            action_table.input2,
            action_table.input3,
            action_table.input4,
            action_table.input5,
            action_table.input6,
            action_table.input7,
            action_table.input8,
        ]

        # Encode each input channel
        for input_index, input_channel in enumerate(input_channels):
            Xp20MsActionTableSerializer._encode_input_channel(
                input_channel, input_index, raw_bytes
            )

        encoded_data = nibbles(raw_bytes)
        # Convert raw bytes to hex string with A-P encoding
        return "AAAA" + encoded_data

    @staticmethod
    def from_data(msactiontable_rawdata: str) -> Xp20MsActionTable:
        """Deserialize telegram data to XP20 action table.

        Args:
            msactiontable_rawdata: 64-character hex string with A-P encoding

        Returns:
            Decoded XP20 action table

        Raises:
            ValueError: If input length is not 64 characters
        """
        raw_length = len(msactiontable_rawdata)
        if raw_length < 68:  # Minimum: 4 char prefix + 64 chars data
            raise ValueError(
                f"XP20 action table data must be 68 characters long, got {len(msactiontable_rawdata)}"
            )

        # Remove action table count prefix (first 4 characters: AAAA, AAAB, etc.)
        data = msactiontable_rawdata[4:]

        # Take first 64 chars (32 bytes) as per pseudocode
        hex_data = data[:64]

        # Convert hex string to bytes using deNibble (A-P encoding)
        raw_bytes = de_nibbles(hex_data)

        # Decode input channels
        input_channels = []
        for input_index in range(8):
            input_channel = Xp20MsActionTableSerializer._decode_input_channel(
                raw_bytes, input_index
            )
            input_channels.append(input_channel)

        # Create and return XP20 action table
        return Xp20MsActionTable(
            input1=input_channels[0],
            input2=input_channels[1],
            input3=input_channels[2],
            input4=input_channels[3],
            input5=input_channels[4],
            input6=input_channels[5],
            input7=input_channels[6],
            input8=input_channels[7],
        )

    @staticmethod
    def _decode_input_channel(raw_bytes: bytearray, input_index: int) -> InputChannel:
        """Extract input channel configuration from raw bytes.

        Args:
            raw_bytes: Raw byte array from telegram
            input_index: Input channel index (0-7)

        Returns:
            Decoded input channel configuration
        """
        # Extract bit flags from appropriate offsets
        short_long_flags = byte_to_bits(raw_bytes[SHORT_LONG_INDEX])
        group_on_off_flags = byte_to_bits(raw_bytes[GROUP_ON_OFF_INDEX])
        invert_flags = byte_to_bits(raw_bytes[INVERT_INDEX])
        sa_function_flags = byte_to_bits(raw_bytes[SA_FUNCTION_INDEX])
        ta_function_flags = byte_to_bits(raw_bytes[TA_FUNCTION_INDEX])

        # Extract AND functions for this input (full byte)
        and_functions_byte = raw_bytes[AND_FUNCTIONS_INDEX + input_index]
        and_functions = byte_to_bits(and_functions_byte)

        # Create and return input channel
        return InputChannel(
            invert=invert_flags[input_index],
            short_long=short_long_flags[input_index],
            group_on_off=group_on_off_flags[input_index],
            and_functions=and_functions,
            sa_function=sa_function_flags[input_index],
            ta_function=ta_function_flags[input_index],
        )

    @staticmethod
    def _encode_input_channel(
        input_channel: InputChannel, input_index: int, raw_bytes: bytearray
    ) -> None:
        """Encode input channel configuration into raw bytes.

        Args:
            input_channel: Input channel configuration to encode
            input_index: Input channel index (0-7)
            raw_bytes: Raw byte array to modify
        """
        # Set bit flags at appropriate positions
        if input_channel.short_long:
            raw_bytes[SHORT_LONG_INDEX] |= 1 << input_index

        if input_channel.group_on_off:
            raw_bytes[GROUP_ON_OFF_INDEX] |= 1 << input_index

        if input_channel.invert:
            raw_bytes[INVERT_INDEX] |= 1 << input_index

        if input_channel.sa_function:
            raw_bytes[SA_FUNCTION_INDEX] |= 1 << input_index

        if input_channel.ta_function:
            raw_bytes[TA_FUNCTION_INDEX] |= 1 << input_index

        # Encode AND functions (ensure we have exactly 8 bits)
        and_functions = input_channel.and_functions or [False] * 8
        and_functions_byte = 0
        for bit_index, bit_value in enumerate(
            and_functions[:8]
        ):  # Take only first 8 bits
            if bit_value:
                and_functions_byte |= 1 << bit_index

        raw_bytes[AND_FUNCTIONS_INDEX + input_index] = and_functions_byte



"""Datapoint type enumeration for system telegrams."""

from enum import Enum
from typing import Optional


class DataPointType(str, Enum):
    """Data point types for system telegrams.

    Attributes:
        MODULE_TYPE: Module type (XP24, XP33, etc).
        HW_VERSION: Hardware version information.
        SW_VERSION: Software version information.
        SERIAL_NUMBER: Serial number.
        LINK_NUMBER: Link number.
        MODULE_NUMBER: Module number.
        SYSTEM_TYPE: System type.
        MODULE_TYPE_CODE: Module type code.
        MODULE_TYPE_ID: Module type ID.
        MODULE_STATE: Module state.
        MODULE_ERROR_CODE: Status query data point.
        MODULE_INPUT_STATE: Module input state.
        MODULE_OUTPUT_STATE: Channel states (XP33).
        MODULE_FW_CRC: Module firmware CRC.
        MODULE_ACTION_TABLE_CRC: Module action table CRC.
        MODULE_LIGHT_LEVEL: Module light level.
        MODULE_OPERATING_HOURS: Module operating hours.
        MODULE_ENERGY_LEVEL: Current data point.
        TEMPERATURE: Temperature data point.
        SW_TOP_VERSION: Software top version.
        VOLTAGE: Voltage data point.
        AUTO_REPORT_STATUS: Auto report status.
    """

    MODULE_TYPE = "00"  # Module type (XP24, XP33, ..)
    HW_VERSION = "01"  # Hardware version information
    SW_VERSION = "02"  # Software version information
    SERIAL_NUMBER = "03"  # Serial number
    LINK_NUMBER = "04"  # Link number
    MODULE_NUMBER = "05"  # Module number
    SYSTEM_TYPE = "06"  # System type
    MODULE_TYPE_CODE = "07"  # Module type code
    MODULE_TYPE_ID = "08"  # Module type id
    MODULE_STATE = "09"  # Module state
    MODULE_ERROR_CODE = "10"  # Status query data point
    MODULE_INPUT_STATE = "11"  # Module input state
    MODULE_OUTPUT_STATE = "12"  # Channel states (XP33)
    MODULE_FW_CRC = "13"  # Module Firmware CRC
    MODULE_ACTION_TABLE_CRC = "14"  # Module Action Table CRC

    # XP24 00:000[%],01:000[%],02:000[%],03:000[%]
    # XP33 00:000[%],01:000[%],02:000[%]
    MODULE_LIGHT_LEVEL = "15"  # Module Light Level

    # XP24 00:000[H],01:000[H],02:000[H],03:000[H]
    MODULE_OPERATING_HOURS = "16"  # Module Operating Hours

    # XP24 00:00000[NA],01:00000[NA],02:00000[NA],03:00000[NA]
    MODULE_ENERGY_LEVEL = "17"  # Current data point

    # XP24 +34,0C
    # XP33 -20,0C
    TEMPERATURE = "18"  # Temperature data point

    SW_TOP_VERSION = "19"  # Software Top Version
    VOLTAGE = "20"  # VOLTAGE data point
    AUTO_REPORT_STATUS = "21"  # Auto Report Status

    @classmethod
    def from_code(cls, code: str) -> Optional["DataPointType"]:
        """Get DataPointType from code string.

        Args:
            code: Datapoint type code string.

        Returns:
            DataPointType instance if found, None otherwise.
        """
        for dp_type in cls:
            if dp_type.value == code:
                return dp_type
        return None



"""Connection layer for XP CLI tool."""

from xp.connection.exceptions import (
    ProtocolError,
    ValidationError,
    XPError,
)

__all__ = [
    "XPError",
    "ProtocolError",
    "ValidationError",
]



"""Status message models for terminal UI."""

from textual.message import Message


class StatusMessageChanged(Message):
    """Message posted when status message changes."""

    def __init__(self, message: str) -> None:
        """Initialize the message.

        Args:
            message: The status message to display.
        """
        super().__init__()
        self.message = message



"""Write config type enumeration."""

from enum import Enum
from typing import Optional


class WriteConfigType(str, Enum):
    """Write Config types for system telegrams.

    Attributes:
        LINK_NUMBER: Link number configuration (code 04).
        MODULE_NUMBER: Module number configuration (code 05).
        SYSTEM_TYPE: System type configuration (code 06).
    """

    LINK_NUMBER = "04"
    MODULE_NUMBER = "05"
    SYSTEM_TYPE = "06"  # 00 CP, 01 XP, 02 MIXED

    @classmethod
    def from_code(cls, code: str) -> Optional["WriteConfigType"]:
        """Get WriteConfigType from code string.

        Args:
            code: Configuration type code string.

        Returns:
            WriteConfigType instance if found, None otherwise.
        """
        for dp_type in cls:
            if dp_type.value == code:
                return dp_type
        return None



"""XP24 Action Table models for input actions and settings."""

from dataclasses import dataclass, field

from xp.models.telegram.input_action_type import InputActionType
from xp.models.telegram.timeparam_type import TimeParam


@dataclass
class InputAction:
    """Represents an input action with type and parameter.

    Attributes:
        type: The input action type.
        param: Time parameter for the action.
    """

    type: InputActionType = InputActionType.TOGGLE
    param: TimeParam = TimeParam.NONE


@dataclass
class Xp24MsActionTable:
    """XP24 Action Table for managing input actions and settings.

    Each input has an action type (TOGGLE, ON, LEVELSET, etc.)
    with an optional parameter string.

    Attributes:
        MS300: Timing constant for 300ms.
        MS500: Timing constant for 500ms.
        input1_action: Action configuration for input 1.
        input2_action: Action configuration for input 2.
        input3_action: Action configuration for input 3.
        input4_action: Action configuration for input 4.
        mutex12: Mutual exclusion between inputs 1-2.
        mutex34: Mutual exclusion between inputs 3-4.
        curtain12: Curtain setting for inputs 1-2.
        curtain34: Curtain setting for inputs 3-4.
        mutual_deadtime: Master timing (MS300=12 or MS500=20).
    """

    # MS timing constants
    MS300 = 12
    MS500 = 20

    # Input actions for each input (default to TOGGLE with None parameter)
    input1_action: InputAction = field(default_factory=InputAction)
    input2_action: InputAction = field(default_factory=InputAction)
    input3_action: InputAction = field(default_factory=InputAction)
    input4_action: InputAction = field(default_factory=InputAction)

    # Boolean settings
    mutex12: bool = False  # Mutual exclusion between inputs 1-2
    mutex34: bool = False  # Mutual exclusion between inputs 3-4
    curtain12: bool = False  # Curtain setting for inputs 1-2
    curtain34: bool = False  # Curtain setting for inputs 3-4
    mutual_deadtime: int = MS300  # Master timing (MS300=12 or MS500=20)



"""XP230 Server Service for device emulation.

This service provides XP230-specific device emulation functionality,
including response generation and device configuration handling.
"""

from typing import Dict, Optional

from xp.models import ModuleTypeCode
from xp.services.actiontable.msactiontable_serializer import MsActionTableSerializer
from xp.services.server.base_server_service import BaseServerService


class XP230ServerError(Exception):
    """Raised when XP230 server operations fail."""

    pass


class XP230ServerService(BaseServerService):
    """
    XP230 device emulation service.

    Generates XP230-specific responses, handles XP230 device configuration,
    and implements XP230 telegram format.
    """

    def __init__(
        self,
        serial_number: str,
        _variant: str = "",
        _msactiontable_serializer: Optional[MsActionTableSerializer] = None,
    ):
        """Initialize XP230 server service.

        Args:
            serial_number: The device serial number.
            _variant: Reserved parameter for consistency (unused).
            _msactiontable_serializer: Generic MsActionTable serializer (unused).
        """
        super().__init__(serial_number)
        self.device_type = "XP230"
        self.module_type_code = ModuleTypeCode.XP230  # XP230 module type from registry
        self.firmware_version = "XP230_V1.00.04"

    def get_device_info(self) -> Dict:
        """Get XP230 device information.

        Returns:
            Dictionary containing device information.
        """
        return {
            "serial_number": self.serial_number,
            "device_type": self.device_type,
            "firmware_version": self.firmware_version,
            "status": self.device_status,
            "link_number": self.link_number,
        }



"""Conbus DataPoint Query All Service.

This module provides service for querying all datapoint types from a module.
"""

import logging
from datetime import datetime
from typing import Callable, Optional

from twisted.internet.posixbase import PosixReactorBase

from xp.models import ConbusClientConfig, ConbusDatapointResponse
from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.models.telegram.datapoint_type import DataPointType
from xp.models.telegram.reply_telegram import ReplyTelegram
from xp.models.telegram.system_function import SystemFunction
from xp.models.telegram.telegram_type import TelegramType
from xp.services import TelegramService
from xp.services.protocol import ConbusProtocol


class ConbusDatapointQueryAllService(ConbusProtocol):
    """
    Utility service for querying all datapoints from a module.

    This service orchestrates multiple ConbusDatapointService calls to query
    all available datapoint types sequentially.
    """

    def __init__(
        self,
        telegram_service: TelegramService,
        cli_config: ConbusClientConfig,
        reactor: PosixReactorBase,
    ) -> None:
        """Initialize the query all service.

        Args:
            telegram_service: TelegramService for dependency injection.
            cli_config: ConbusClientConfig for connection settings.
            reactor: PosixReactorBase for async operations.
        """
        super().__init__(cli_config, reactor)
        self.telegram_service = telegram_service
        self.serial_number: str = ""
        self.finish_callback: Optional[Callable[[ConbusDatapointResponse], None]] = None
        self.progress_callback: Optional[Callable[[ReplyTelegram], None]] = None
        self.service_response: ConbusDatapointResponse = ConbusDatapointResponse(
            success=False,
            serial_number=self.serial_number,
        )
        self.datapoint_types = list(DataPointType)
        self.current_index = 0

        # Set up logging
        self.logger = logging.getLogger(__name__)

    def connection_established(self) -> None:
        """Handle connection established event."""
        self.logger.debug("Connection established, querying datapoints.")
        self.next_datapoint()

    def next_datapoint(self) -> bool:
        """Query the next datapoint type.

        Returns:
            True if there are more datapoints to query, False otherwise.
        """
        self.logger.debug("Querying next datapoint")

        if self.current_index >= len(self.datapoint_types):
            return False

        datapoint_type_code = self.datapoint_types[self.current_index]
        datapoint_type = DataPointType(datapoint_type_code)

        self.logger.debug(f"Datapoint: {datapoint_type}")
        self.send_telegram(
            telegram_type=TelegramType.SYSTEM,
            serial_number=self.serial_number,
            system_function=SystemFunction.READ_DATAPOINT,
            data_value=str(datapoint_type.value),
        )
        self.current_index += 1
        return True

    def timeout(self) -> bool:
        """Handle timeout event by querying next datapoint.

        Returns:
            True to continue, False to stop the reactor.
        """
        self.logger.debug("Timeout, querying next datapoint")
        query_next_datapoint = self.next_datapoint()
        if not query_next_datapoint:
            if self.finish_callback:
                self.logger.debug("Received all datapoints telegram")
                self.service_response.success = True
                self.service_response.timestamp = datetime.now()
                self.service_response.serial_number = self.serial_number
                self.service_response.system_function = SystemFunction.READ_DATAPOINT
                self.finish_callback(self.service_response)
                return False
        return True

    def telegram_sent(self, telegram_sent: str) -> None:
        """Handle telegram sent event.

        Args:
            telegram_sent: The telegram that was sent.
        """
        self.service_response.sent_telegram = telegram_sent

    def telegram_received(self, telegram_received: TelegramReceivedEvent) -> None:
        """Handle telegram received event.

        Args:
            telegram_received: The telegram received event.
        """
        self.logger.debug(f"Telegram received: {telegram_received}")
        if not self.service_response.received_telegrams:
            self.service_response.received_telegrams = []
        self.service_response.received_telegrams.append(telegram_received.frame)

        if (
            not telegram_received.checksum_valid
            or telegram_received.telegram_type != TelegramType.REPLY
            or telegram_received.serial_number != self.serial_number
        ):
            self.logger.debug("Not a reply for our serial number")
            return

        # Parse the reply telegram
        datapoint_telegram = self.telegram_service.parse_reply_telegram(
            telegram_received.frame
        )
        if (
            not datapoint_telegram
            or datapoint_telegram.system_function != SystemFunction.READ_DATAPOINT
        ):
            self.logger.debug("Not a reply for our datapoint type")
            return

        self.logger.debug("Received a datapoint telegram")
        if self.progress_callback:
            self.progress_callback(datapoint_telegram)

    def failed(self, message: str) -> None:
        """Handle failed connection event.

        Args:
            message: Failure message.
        """
        self.logger.debug(f"Failed with message: {message}")
        self.service_response.success = False
        self.service_response.timestamp = datetime.now()
        self.service_response.error = message
        if self.finish_callback:
            self.finish_callback(self.service_response)

    def query_all_datapoints(
        self,
        serial_number: str,
        finish_callback: Callable[[ConbusDatapointResponse], None],
        progress_callback: Callable[[ReplyTelegram], None],
        timeout_seconds: Optional[float] = None,
    ) -> None:
        """Query all datapoints from a module.

        Args:
            serial_number: 10-digit module serial number.
            finish_callback: Callback function to call when all datapoints are received.
            progress_callback: Callback function to call when each datapoint is received.
            timeout_seconds: Timeout in seconds.
        """
        self.logger.info("Starting query_all_datapoints")
        if timeout_seconds:
            self.timeout_seconds = timeout_seconds
        self.finish_callback = finish_callback
        self.progress_callback = progress_callback
        self.serial_number = serial_number
        self.start_reactor()



"""TUI widgets package."""

from xp.term.widgets.help_menu import HelpMenuWidget
from xp.term.widgets.protocol_log import ProtocolLogWidget
from xp.term.widgets.status_footer import StatusFooterWidget

__all__ = ["HelpMenuWidget", "ProtocolLogWidget", "StatusFooterWidget"]



"""Conson Configuration Validator for HomeKit.

This module validates conson.yml configuration files for HomeKit integration.
"""

from typing import List, Set

from xp.models.homekit.homekit_conson_config import (
    ConsonModuleConfig,
    ConsonModuleListConfig,
)


class ConsonConfigValidator:
    """Validates conson.yml configuration file for HomeKit integration."""

    def __init__(self, config: ConsonModuleListConfig):
        """Initialize the Conson config validator.

        Args:
            config: Conson module list configuration to validate.
        """
        self.config = config

    def validate_unique_names(self) -> List[str]:
        """Validate that all module names are unique.

        Returns:
            List of validation error messages.
        """
        names: Set[str] = set()
        errors = []

        for module in self.config.root:
            if module.name in names:
                errors.append(f"Duplicate module name: {module.name}")
            names.add(module.name)

        return errors

    def validate_unique_serial_numbers(self) -> List[str]:
        """Validate that all serial numbers are unique.

        Returns:
            List of validation error messages.
        """
        serials: Set[str] = set()
        errors = []

        for module in self.config.root:
            if module.serial_number in serials:
                errors.append(f"Duplicate serial number: {module.serial_number}")
            serials.add(module.serial_number)

        return errors

    def validate_module_type_codes(self) -> List[str]:
        """Validate module type code ranges.

        Returns:
            List of validation error messages.
        """
        errors = [
            f"Invalid module_type_code {module.module_type_code} for module {module.name}. Must be between 1 and 255."
            for module in self.config.root
            if not (1 <= module.module_type_code <= 255)
        ]

        return errors

    def validate_network_config(self) -> List[str]:
        """Validate IP/port configuration.

        Returns:
            List of validation error messages.
        """
        errors = [
            f"Invalid conbus_port {module.conbus_port} for module {module.name}. Must be between 1 and 65535."
            for module in self.config.root
            if module.conbus_port is not None and not (1 <= module.conbus_port <= 65535)
        ]

        return errors

    def validate_all(self) -> List[str]:
        """Run all validations and return combined errors.

        Returns:
            List of all validation error messages.
        """
        all_errors = []
        all_errors.extend(self.validate_unique_names())
        all_errors.extend(self.validate_unique_serial_numbers())
        all_errors.extend(self.validate_module_type_codes())
        all_errors.extend(self.validate_network_config())
        return all_errors

    def get_module_by_serial(self, serial_number: str) -> ConsonModuleConfig:
        """Get module configuration by serial number.

        Args:
            serial_number: Serial number of the module to find.

        Returns:
            Module configuration for the specified serial number.

        Raises:
            ValueError: If module with serial number is not found.
        """
        for module in self.config.root:
            if module.serial_number == serial_number:
                return module
        raise ValueError(f"Module with serial number {serial_number} not found")

    def get_all_serial_numbers(self) -> Set[str]:
        """Get all serial numbers from the configuration.

        Returns:
            Set of all serial numbers in the configuration.
        """
        return {module.serial_number for module in self.config.root}



"""Conbus Blink Service for TCP communication with Conbus servers.

This service implements a TCP client that connects to Conbus servers and sends
blink/unblink telegrams to control module LED indicators.
"""

import logging
from datetime import datetime
from typing import Callable, Optional

from twisted.internet.posixbase import PosixReactorBase

from xp.models import ConbusClientConfig
from xp.models.conbus.conbus_blink import ConbusBlinkResponse
from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.models.telegram.system_function import SystemFunction
from xp.models.telegram.telegram_type import TelegramType
from xp.services.protocol import ConbusProtocol
from xp.services.telegram.telegram_service import TelegramService


class ConbusBlinkService(ConbusProtocol):
    """
    Service for blinking module LEDs on Conbus servers.

    Uses ConbusProtocol to provide blink/unblink functionality
    for controlling module LED indicators.
    """

    def __init__(
        self,
        telegram_service: TelegramService,
        cli_config: ConbusClientConfig,
        reactor: PosixReactorBase,
    ) -> None:
        """Initialize the Conbus blink service.

        Args:
            telegram_service: Service for parsing telegrams.
            cli_config: Configuration for Conbus client connection.
            reactor: Twisted reactor for event loop.
        """
        super().__init__(cli_config, reactor)
        self.telegram_service = telegram_service
        self.serial_number: str = ""
        self.on_or_off = "none"
        self.finish_callback: Optional[Callable[[ConbusBlinkResponse], None]] = None
        self.service_response: ConbusBlinkResponse = ConbusBlinkResponse(
            success=False,
            serial_number=self.serial_number,
            system_function=SystemFunction.NONE,
            operation=self.on_or_off,
        )

        # Set up logging
        self.logger = logging.getLogger(__name__)

    def connection_established(self) -> None:
        """Handle connection established event."""
        self.logger.debug("Connection established, sending blink command.")
        # Blink is 05, Unblink is 06
        system_function = SystemFunction.UNBLINK
        if self.on_or_off.lower() == "on":
            system_function = SystemFunction.BLINK

        self.send_telegram(
            telegram_type=TelegramType.SYSTEM,
            serial_number=self.serial_number,
            system_function=system_function,
            data_value="00",
        )
        self.service_response.system_function = system_function
        self.service_response.operation = self.on_or_off

    def telegram_sent(self, telegram_sent: str) -> None:
        """Handle telegram sent event.

        Args:
            telegram_sent: The telegram that was sent.
        """
        system_telegram = self.telegram_service.parse_system_telegram(telegram_sent)
        self.service_response.sent_telegram = system_telegram

    def telegram_received(self, telegram_received: TelegramReceivedEvent) -> None:
        """Handle telegram received event.

        Args:
            telegram_received: The telegram received event.
        """
        self.logger.debug(f"Telegram received: {telegram_received}")
        if not self.service_response.received_telegrams:
            self.service_response.received_telegrams = []
        self.service_response.received_telegrams.append(telegram_received.frame)

        if (
            not telegram_received.checksum_valid
            or telegram_received.telegram_type != TelegramType.REPLY
            or telegram_received.serial_number != self.serial_number
        ):
            self.logger.debug("Not a reply")
            return

        reply_telegram = self.telegram_service.parse_reply_telegram(
            telegram_received.frame
        )
        if reply_telegram is not None and reply_telegram.system_function in (
            SystemFunction.ACK,
            SystemFunction.NAK,
        ):
            self.logger.debug("Received blink response")
            self.service_response.success = True
            self.service_response.timestamp = datetime.now()
            self.service_response.serial_number = self.serial_number
            self.service_response.reply_telegram = reply_telegram

            if self.finish_callback:
                self.finish_callback(self.service_response)

    def failed(self, message: str) -> None:
        """Handle failed connection event.

        Args:
            message: Failure message.
        """
        self.logger.debug(f"Failed with message: {message}")
        self.service_response.success = False
        self.service_response.timestamp = datetime.now()
        self.service_response.error = message
        if self.finish_callback:
            self.finish_callback(self.service_response)

    def send_blink_telegram(
        self,
        serial_number: str,
        on_or_off: str,
        finish_callback: Callable[[ConbusBlinkResponse], None],
        timeout_seconds: Optional[float] = None,
    ) -> None:
        r"""Send blink command to start blinking module LED.

        Args:
            serial_number: 10-digit module serial number.
            on_or_off: "on" to blink or "off" to unblink.
            finish_callback: Callback function to call when the reply is received.
            timeout_seconds: Timeout in seconds.

        Examples:
            \b
            xp conbus blink 0012345008 on
            xp conbus blink 0012345008 off
        """
        self.logger.info("Starting send_blink_telegram")
        if timeout_seconds:
            self.timeout_seconds = timeout_seconds
        self.finish_callback = finish_callback
        self.serial_number = serial_number
        self.on_or_off = on_or_off
        self.start_reactor()



"""Term CLI group definition for TUI commands."""

import click
from click_help_colors import HelpColorsGroup


@click.group(
    cls=HelpColorsGroup, help_headers_color="yellow", help_options_color="green"
)
def term() -> None:
    """Terminal UI commands for interactive monitoring and control."""
    pass



"""HomeKit Conbus Service for protocol communication.

This module bridges HomeKit events with the Conbus protocol for device control.
"""

import logging

from bubus import EventBus

from xp.models.protocol.conbus_protocol import (
    ReadDatapointFromProtocolEvent,
    SendActionEvent,
    SendWriteConfigEvent,
)
from xp.models.telegram.datapoint_type import DataPointType
from xp.models.telegram.system_function import SystemFunction
from xp.services.protocol.telegram_protocol import TelegramProtocol


class HomeKitConbusService:
    """Service for bridging HomeKit events with Conbus protocol.

    Attributes:
        event_bus: Event bus for inter-service communication.
        telegram_protocol: Protocol for sending telegrams.
        logger: Logger instance.
    """

    event_bus: EventBus

    def __init__(self, event_bus: EventBus, telegram_protocol: TelegramProtocol):
        """Initialize the HomeKit Conbus service.

        Args:
            event_bus: Event bus instance.
            telegram_protocol: Telegram protocol instance.
        """
        self.logger = logging.getLogger(__name__)
        self.event_bus = event_bus
        self.telegram_protocol = telegram_protocol

        # Register event handlers
        self.event_bus.on(
            ReadDatapointFromProtocolEvent, self.handle_read_datapoint_request
        )
        self.event_bus.on(SendActionEvent, self.handle_send_action_event)
        self.event_bus.on(SendWriteConfigEvent, self.handle_send_write_config_event)

    def handle_read_datapoint_request(
        self, event: ReadDatapointFromProtocolEvent
    ) -> None:
        """Handle request to read datapoint from protocol.

        Args:
            event: Read datapoint event with serial number and datapoint type.
        """
        self.logger.debug(f"read_datapoint_request {event}")

        system_function = SystemFunction.READ_DATAPOINT.value
        datapoint_value = event.datapoint_type.value
        telegram = f"S{event.serial_number}F{system_function}D{datapoint_value}"
        self.telegram_protocol.sendFrame(telegram.encode())

    def handle_send_write_config_event(self, event: SendWriteConfigEvent) -> None:
        """Handle send write config event.

        Args:
            event: Write config event with configuration data.
        """
        self.logger.debug(f"send_write_config_event {event}")

        # Format data as output_number:level (e.g., "02:050")
        system_function = SystemFunction.WRITE_CONFIG.value
        datapoint_type = DataPointType.MODULE_LIGHT_LEVEL.value
        config_data = f"{event.output_number:02d}:{event.value:03d}"
        telegram = (
            f"S{event.serial_number}F{system_function}D{datapoint_type}{config_data}"
        )
        self.telegram_protocol.sendFrame(telegram.encode())

    def handle_send_action_event(self, event: SendActionEvent) -> None:
        """Handle send action event.

        Args:
            event: Send action event with action data.
        """
        self.logger.debug(f"send_action_event {event}")

        telegram = event.on_action if event.value else event.off_action
        telegram_make = f"{telegram}M"
        telegram_break = f"{telegram}B"
        self.telegram_protocol.sendFrame(telegram_make.encode())
        self.telegram_protocol.sendFrame(telegram_break.encode())



"""Conbus client operations CLI commands."""

import json

import click

from xp.cli.commands.conbus.conbus import conbus_output
from xp.cli.utils.decorators import (
    connection_command,
)
from xp.cli.utils.serial_number_type import SERIAL
from xp.models import ConbusDatapointResponse
from xp.models.conbus.conbus_output import ConbusOutputResponse
from xp.models.telegram.action_type import ActionType
from xp.models.telegram.datapoint_type import DataPointType
from xp.services.conbus.conbus_datapoint_service import ConbusDatapointService
from xp.services.conbus.conbus_output_service import ConbusOutputService


@conbus_output.command("on")
@click.argument("serial_number", type=SERIAL)
@click.argument("output_number", type=int)
@click.pass_context
@connection_command()
def xp_output_on(ctx: click.Context, serial_number: str, output_number: int) -> None:
    r"""Send ON command for output_number XP module serial_number.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.
        output_number: Output number.

    Examples:
        \b
        xp conbus output on 0011223344 0  # Turn on output 0
    """
    service: ConbusOutputService = (
        ctx.obj.get("container").get_container().resolve(ConbusOutputService)
    )

    def on_finish(response: ConbusOutputResponse) -> None:
        """Handle successful completion of output on command.

        Args:
            response: Output response object.
        """
        click.echo(json.dumps(response.to_dict(), indent=2))

    with service:
        service.send_action(
            serial_number=serial_number,
            output_number=output_number,
            action_type=ActionType.ON_RELEASE,
            finish_callback=on_finish,
        )


@conbus_output.command("off")
@click.argument("serial_number", type=SERIAL)
@click.argument("output_number", type=int)
@click.pass_context
@connection_command()
def xp_output_off(ctx: click.Context, serial_number: str, output_number: int) -> None:
    r"""Send OFF command for output_number XP module serial_number.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.
        output_number: Output number.

    Examples:
        \b
        xp conbus output off 0011223344 1    # Turn off output 1
    """
    service: ConbusOutputService = (
        ctx.obj.get("container").get_container().resolve(ConbusOutputService)
    )

    def on_finish(response: ConbusOutputResponse) -> None:
        """Handle successful completion of output off command.

        Args:
            response: Output response object.
        """
        click.echo(json.dumps(response.to_dict(), indent=2))

    with service:
        service.send_action(
            serial_number=serial_number,
            output_number=output_number,
            action_type=ActionType.OFF_PRESS,
            finish_callback=on_finish,
        )


@conbus_output.command("status")
@click.argument("serial_number", type=SERIAL)
@click.pass_context
@connection_command()
def xp_output_status(ctx: click.Context, serial_number: str) -> None:
    r"""Query output state command to XP module serial_number.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.

    Examples:
        \b
        xp conbus output status 0011223344    # Query output status
    """
    service: ConbusDatapointService = (
        ctx.obj.get("container").get_container().resolve(ConbusDatapointService)
    )

    def on_finish(response: ConbusDatapointResponse) -> None:
        """Handle successful completion of output status query.

        Args:
            response: Datapoint response object.
        """
        click.echo(json.dumps(response.to_dict(), indent=2))

    with service:
        service.query_datapoint(
            serial_number=serial_number,
            datapoint_type=DataPointType.MODULE_OUTPUT_STATE,
            finish_callback=on_finish,
        )


@conbus_output.command("state")
@click.argument("serial_number", type=SERIAL)
@click.pass_context
@connection_command()
def xp_module_state(ctx: click.Context, serial_number: str) -> None:
    r"""Query module state of the XP module serial_number.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.

    Examples:
        \b
        xp conbus output state 0011223344    # Query module state
    """
    service: ConbusDatapointService = (
        ctx.obj.get("container").get_container().resolve(ConbusDatapointService)
    )

    def on_finish(response: ConbusDatapointResponse) -> None:
        """Handle successful completion of module state query.

        Args:
            response: Datapoint response object.
        """
        click.echo(json.dumps(response.to_dict(), indent=2))

    with service:
        service.query_datapoint(
            serial_number=serial_number,
            datapoint_type=DataPointType.MODULE_STATE,
            finish_callback=on_finish,
        )



"""HomeKit HAP Service for Apple HomeKit integration.

This module provides the main HAP (HomeKit Accessory Protocol) service.
"""

import logging
import signal
import threading
from datetime import datetime
from typing import Dict, List, Optional

from bubus import EventBus
from pyhap.accessory import Bridge
from pyhap.accessory_driver import AccessoryDriver
from typing_extensions import Union

import xp
from xp.models.homekit.homekit_accessory import TemperatureSensor
from xp.models.homekit.homekit_config import (
    HomekitAccessoryConfig,
    HomekitConfig,
    RoomConfig,
)
from xp.models.protocol.conbus_protocol import (
    LightLevelReceivedEvent,
    ModuleStateChangedEvent,
    OutputStateReceivedEvent,
    ReadDatapointEvent,
)
from xp.models.telegram.datapoint_type import DataPointType
from xp.services.homekit.homekit_dimminglight import DimmingLight
from xp.services.homekit.homekit_lightbulb import LightBulb
from xp.services.homekit.homekit_module_service import HomekitModuleService
from xp.services.homekit.homekit_outlet import Outlet


class HomekitHapService:
    """HomeKit HAP service.

    Manages HAP accessory protocol, handles bridge and accessory setup,
    and processes HomeKit events for device state synchronization.

    Attributes:
        event_bus: Event bus for inter-service communication.
        last_activity: Timestamp of last service activity.
        logger: Logger instance.
        config: HomeKit configuration.
        accessory_registry: Registry of accessories by identifier.
        module_registry: Registry of accessories by module key.
        modules: Module service for module lookup.
        driver: HAP accessory driver.
    """

    event_bus: EventBus

    def __init__(
        self,
        homekit_config: HomekitConfig,
        module_service: HomekitModuleService,
        event_bus: EventBus,
    ):
        """Initialize the HomeKit HAP service.

        Args:
            homekit_config: HomeKit configuration.
            module_service: Module service for dependency injection.
            event_bus: Event bus for dependency injection.
        """
        self.last_activity: Optional[datetime] = None

        # Set up logging
        self.logger = logging.getLogger(__name__)

        # Load configuration
        self.config = homekit_config
        self.accessory_registry: Dict[str, Union[LightBulb, Outlet, DimmingLight]] = {}
        self.module_registry: Dict[
            tuple[int, int], List[Union[LightBulb, Outlet, DimmingLight]]
        ] = {}

        # Service dependencies
        self.modules = module_service
        self.event_bus = event_bus

        # Subscribe to events
        self.event_bus.on(ModuleStateChangedEvent, self.handle_module_state_changed)
        self.event_bus.on(OutputStateReceivedEvent, self.handle_output_state_received)
        self.event_bus.on(LightLevelReceivedEvent, self.handle_light_level_received)

        # We want SIGTERM (terminate) to be handled by the driver itself,
        # so that it can gracefully stop the accessory, server and advertising.
        driver = AccessoryDriver(
            port=self.config.homekit.port,
        )
        signal.signal(signal.SIGTERM, driver.signal_handler)
        self.driver: AccessoryDriver = driver

    async def async_start(self) -> None:
        """Start the HAP service asynchronously."""
        self.logger.info("Loading accessories.")
        self.build_bridge()
        self.logger.info("Accessories loaded successfully")

        # Start HAP-python in a separate thread to avoid event loop conflicts
        self.logger.info("Starting HAP-python driver in separate thread.")
        hap_thread = threading.Thread(
            target=self._run_driver_in_thread, daemon=True, name="HAP-Python"
        )
        hap_thread.start()
        self.logger.info("HAP-python driver thread started")

    def _run_driver_in_thread(self) -> None:
        """Run the HAP-python driver in a separate thread with its own event loop."""
        try:
            self.logger.info("HAP-python thread starting, creating new event loop.")
            # Create a new event loop for this thread

            self.logger.info("Starting HAP-python driver.")
            self.driver.start()
            self.logger.info("HAP-python driver started successfully")
        except Exception as e:
            self.logger.error(f"HAP-python driver error: {e}", exc_info=True)

    def handle_output_state_received(self, event: OutputStateReceivedEvent) -> str:
        """Handle output state received event.

        Args:
            event: Output state received event.

        Returns:
            Data value from the event.
        """
        self.logger.debug(f"Received OutputStateReceivedEvent {event}")
        output_number = 0
        for output in event.data_value[::-1]:
            if output == "x":
                break
            identifier = f"{event.serial_number}.{output_number:02X}"
            accessory = self.accessory_registry.get(identifier)

            if not accessory:
                self.logger.warning(f"Invalid accessory: {identifier} (not found)")
            else:
                accessory.is_on = True if output == "1" else False
            output_number += 1

        self.logger.debug(
            f"handle_output_state_received "
            f"serial_number: {event.serial_number}, "
            f"data_vale: {event.data_value}"
        )
        return event.data_value

    def handle_light_level_received(self, event: LightLevelReceivedEvent) -> str:
        """Handle light level received event.

        Args:
            event: Light level received event.

        Returns:
            Data value from the event.
        """
        # Parse response format like "00:050,01:025,02:100"
        self.logger.debug("Received LightLevelReceivedEvent", extra={"event": event})
        output_number = 0
        for output_data in event.data_value.split(","):
            if ":" in output_data:
                output_str, level_str = output_data.split(":")
                level_str = level_str.replace("[%]", "")
                output_number = int(output_str)
                brightness = int(level_str)
                identifier = f"{event.serial_number}.{output_number:02X}"
                accessory = self.accessory_registry.get(identifier)

                if not accessory:
                    self.logger.warning(
                        f"Invalid accessory: {event.serial_number} (not found)"
                    )
                elif not isinstance(accessory, DimmingLight):
                    self.logger.warning(
                        f"Invalid accessory: {event.serial_number} (not dimming light)"
                    )
                else:
                    accessory.brightness = brightness

            output_number += 1

        self.logger.debug(
            f"handle_light_level_received "
            f"serial_number: {event.serial_number}, "
            f"data_vale: {event.data_value}"
        )
        return event.data_value

    def build_bridge(self) -> None:
        """Build the HomeKit bridge with all configured accessories."""
        bridge_config = self.config.bridge
        bridge = Bridge(self.driver, bridge_config.name)
        bridge.set_info_service(
            xp.__version__, xp.__manufacturer__, xp.__model__, xp.__serial__
        )

        for room in bridge_config.rooms:
            self.add_room(bridge, room)

        self.driver.add_accessory(accessory=bridge)

    def add_room(self, bridge: Bridge, room: RoomConfig) -> None:
        """Add a room with its accessories to the bridge.

        Args:
            bridge: HAP bridge instance.
            room: Room configuration.
        """
        temperature = TemperatureSensor(self.driver, room.name)
        bridge.add_accessory(temperature)

        for accessory_name in room.accessories:
            homekit_accessory = self.get_accessory_by_name(accessory_name)
            if homekit_accessory is None:
                self.logger.warning("Accessory '{}' not found".format(accessory_name))
                continue

            accessory = self.get_accessory(homekit_accessory)
            if accessory:
                bridge.add_accessory(accessory)
                # Add to accessory_registry
                self.accessory_registry[accessory.identifier] = accessory

                # Add to module_registry for event-driven lookup
                module_key = (
                    accessory.module.module_type_code,
                    accessory.module.link_number,
                )
                if module_key not in self.module_registry:
                    self.module_registry[module_key] = []
                self.module_registry[module_key].append(accessory)

    def get_accessory(
        self, homekit_accessory: HomekitAccessoryConfig
    ) -> Union[LightBulb, Outlet, DimmingLight, None]:
        """Get an accessory instance from configuration.

        Args:
            homekit_accessory: HomeKit accessory configuration.

        Returns:
            Accessory instance or None if not found or invalid service type.
        """
        module_config = self.modules.get_module_by_serial(
            homekit_accessory.serial_number
        )
        if module_config is None:
            self.logger.warning(f"Accessory '{homekit_accessory.name}' not found")
            return None

        if homekit_accessory.service == "lightbulb":
            return LightBulb(
                driver=self.driver,
                module=module_config,
                accessory=homekit_accessory,
                event_bus=self.event_bus,
            )

        if homekit_accessory.service == "outlet":
            return Outlet(
                driver=self.driver,
                module=module_config,
                accessory=homekit_accessory,
                event_bus=self.event_bus,
            )

        if homekit_accessory.service == "dimminglight":
            return DimmingLight(
                driver=self.driver,
                module=module_config,
                accessory=homekit_accessory,
                event_bus=self.event_bus,
            )

        self.logger.warning(f"Accessory '{homekit_accessory.name}' not found")
        return None

    def get_accessory_by_name(self, name: str) -> Optional[HomekitAccessoryConfig]:
        """Get an accessory configuration by name.

        Args:
            name: Name of the accessory to find.

        Returns:
            Accessory configuration if found, None otherwise.
        """
        return next(
            (module for module in self.config.accessories if module.name == name), None
        )

    def handle_module_state_changed(self, event: ModuleStateChangedEvent) -> None:
        """Handle module state change by refreshing affected accessories.

        Args:
            event: Module state changed event.
        """
        self.logger.debug(
            f"Module state changed: module_type={event.module_type_code}, "
            f"link={event.link_number}, input={event.input_number}"
        )

        # O(1) lookup using module_registry
        module_key = (event.module_type_code, event.link_number)
        affected_accessories = self.module_registry.get(module_key, [])

        if not affected_accessories:
            self.logger.debug(
                f"No accessories found for module_type={event.module_type_code}, "
                f"link={event.link_number}"
            )
            return

        # Request cache refresh for each affected accessory
        for accessory in affected_accessories:
            self.logger.info(
                f"Requesting cache refresh for accessory: {accessory.identifier}"
            )

            # Request OUTPUT_STATE refresh
            self.event_bus.dispatch(
                ReadDatapointEvent(
                    serial_number=accessory.module.serial_number,
                    datapoint_type=DataPointType.MODULE_OUTPUT_STATE,
                    refresh_cache=True,
                )
            )

            # If dimming light, also refresh LIGHT_LEVEL
            if isinstance(accessory, DimmingLight):
                self.event_bus.dispatch(
                    ReadDatapointEvent(
                        serial_number=accessory.module.serial_number,
                        datapoint_type=DataPointType.MODULE_LIGHT_LEVEL,
                        refresh_cache=True,
                    )
                )



"""Protocol Factory for Twisted protocol creation.

This module provides factory classes for protocol instantiation.
"""

import logging

from bubus import EventBus
from twisted.internet import protocol
from twisted.internet.interfaces import IAddress, IConnector
from twisted.python.failure import Failure

from xp.models.protocol.conbus_protocol import (
    ConnectionFailedEvent,
    ConnectionLostEvent,
)
from xp.services.protocol import TelegramProtocol


class TelegramFactory(protocol.ClientFactory):
    """Factory for creating Telegram protocol instances.

    Attributes:
        event_bus: Event bus for dispatching protocol events.
        telegram_protocol: Protocol instance to use.
        connector: Connection connector instance.
        logger: Logger instance for this factory.
    """

    def __init__(
        self,
        event_bus: EventBus,
        telegram_protocol: TelegramProtocol,
        connector: IConnector,
    ) -> None:
        """Initialize TelegramFactory.

        Args:
            event_bus: Event bus for protocol events.
            telegram_protocol: Protocol instance to use for connections.
            connector: Connection connector for managing connections.
        """
        self.event_bus = event_bus
        self.telegram_protocol = telegram_protocol
        self.connector = connector
        self.logger = logging.getLogger(__name__)

    def buildProtocol(self, addr: IAddress) -> TelegramProtocol:
        """Build protocol instance for connection.

        Args:
            addr: Address of the connection.

        Returns:
            Telegram protocol instance for this connection.
        """
        self.logger.debug(f"buildProtocol: {addr}")
        return self.telegram_protocol

    def clientConnectionFailed(self, connector: IConnector, reason: Failure) -> None:
        """Handle connection failure event.

        Args:
            connector: Connection connector instance.
            reason: Failure reason details.
        """
        self.event_bus.dispatch(ConnectionFailedEvent(reason=str(reason)))
        self.connector.stop()

    def clientConnectionLost(self, connector: IConnector, reason: Failure) -> None:
        """Handle connection lost event.

        Args:
            connector: Connection connector instance.
            reason: Reason for connection loss.
        """
        self.event_bus.dispatch(ConnectionLostEvent(reason=str(reason)))
        self.connector.stop()



"""Unit tests for ProtocolMonitorService."""

from unittest.mock import Mock

import pytest
from twisted.python.failure import Failure

from xp.models.term.connection_state import ConnectionState
from xp.models.term.protocol_keys_config import ProtocolKeyConfig, ProtocolKeysConfig
from xp.services.term.protocol_monitor_service import ProtocolMonitorService


class TestProtocolMonitorService:
    """Unit tests for ProtocolMonitorService."""

    @pytest.fixture
    def mock_protocol(self):
        """Create mock ConbusEventProtocol."""
        protocol = Mock()
        protocol.cli_config = Mock()
        protocol.cli_config.ip = "192.168.1.100"
        protocol.cli_config.port = 10001
        protocol.transport = None
        protocol.on_connection_made = Mock()
        protocol.on_connection_failed = Mock()
        protocol.on_telegram_received = Mock()
        protocol.on_telegram_sent = Mock()
        protocol.on_timeout = Mock()
        protocol.on_failed = Mock()
        protocol.connect = Mock()
        protocol.disconnect = Mock()
        protocol.send_raw_telegram = Mock()
        return protocol

    @pytest.fixture
    def protocol_keys(self):
        """Create protocol keys config."""
        return ProtocolKeysConfig(
            protocol={
                "1": ProtocolKeyConfig(
                    name="Discover", telegrams=["S0000000000F01D00"]
                ),
                "2": ProtocolKeyConfig(name="Blink", telegrams=["B01"]),
            }
        )

    @pytest.fixture
    def service(self, mock_protocol, protocol_keys):
        """Create service instance."""
        return ProtocolMonitorService(
            conbus_protocol=mock_protocol, protocol_keys=protocol_keys
        )

    def test_initialization(self, service, mock_protocol):
        """Test service initialization connects signals and sets initial state."""
        assert service.connection_state == ConnectionState.DISCONNECTED
        assert service.server_info == "192.168.1.100:10001"
        mock_protocol.on_connection_made.connect.assert_called_once()
        mock_protocol.on_connection_failed.connect.assert_called_once()

    def test_toggle_connection_when_disconnected(self, service, mock_protocol):
        """Test toggle_connection calls _connect when disconnected."""
        service.toggle_connection()
        mock_protocol.connect.assert_called_once()
        assert service.connection_state == ConnectionState.CONNECTING

    def test_toggle_connection_when_connected(self, service, mock_protocol):
        """Test toggle_connection calls _disconnect when connected."""
        # First connect properly through state machine
        service.toggle_connection()  # DISCONNECTED -> CONNECTING
        service._on_connection_made()  # CONNECTING -> CONNECTED

        # Now toggle should disconnect
        service.toggle_connection()
        mock_protocol.disconnect.assert_called_once()

    def test_handle_key_press_valid_key(self, service, mock_protocol):
        """Test handle_key_press sends telegrams for valid keys."""
        result = service.handle_key_press("1")

        assert result is True
        mock_protocol.send_raw_telegram.assert_called_once_with("S0000000000F01D00")

    def test_handle_key_press_invalid_key(self, service):
        """Test handle_key_press returns False for invalid keys."""
        result = service.handle_key_press("invalid")
        assert result is False

    def test_on_connection_made_handler(self, service):
        """Test _on_connection_made updates state and emits signals."""
        # Must be in CONNECTING state first for transition to work
        service._connect()
        service._on_connection_made()

        assert service.connection_state == ConnectionState.CONNECTED

    def test_on_connection_failed_handler(self, service):
        """Test _on_connection_failed updates state."""
        failure = Failure(Exception("Connection error"))
        service._on_connection_failed(failure)

        assert service.connection_state == ConnectionState.DISCONNECTED

    def test_on_telegram_sent_handler(self, service):
        """Test _on_telegram_sent emits display event."""
        service._on_telegram_sent("<S0000000000F01D00FA>")
        # Just verify it doesn't crash

    def test_on_timeout_handler(self, service):
        """Test _on_timeout logs debug message."""
        service._on_timeout()
        # Just verify it doesn't crash

    def test_on_failed_handler(self, service):
        """Test _on_failed updates state."""
        # Must be in valid state for transition to FAILED
        service._connect()  # DISCONNECTED -> CONNECTING
        service._on_failed("Connection lost")
        assert service.connection_state == ConnectionState.FAILED

    def test_get_keys(self, service):
        """Test get_keys returns protocol keys."""
        keys = dict(service.get_keys())
        assert "1" in keys
        assert "2" in keys
        assert keys["1"].name == "Discover"

    def test_context_manager(self, service, mock_protocol):
        """Test context manager entry and exit."""
        with service as svc:
            assert svc is service

        # Verify cleanup was called
        mock_protocol.on_connection_made.disconnect.assert_called_once()

    def test_send_telegram_error_handling(self, service, mock_protocol):
        """Test _send_telegram handles exceptions."""
        mock_protocol.send_raw_telegram.side_effect = Exception("Send failed")

        service._send_telegram("Test", "TELEGRAM")
        # Should not raise, just log error

    def test_cleanup_with_transport(self, service, mock_protocol):
        """Test cleanup disconnects when transport exists."""
        mock_protocol.transport = Mock()
        # Set up proper connected state through state machine
        service._connect()
        service._on_connection_made()

        service.cleanup()

        mock_protocol.disconnect.assert_called_once()
        mock_protocol.on_connection_made.disconnect.assert_called_once()
